import logging
import os
import shutil
import tempfile
import textwrap
import uuid

import jinja2
import pytest

import salt.config
import salt.state
import salt.utils.files
from salt.template import compile_template
from salt.utils.odict import OrderedDict
from salt.utils.pyobjects import (
    DuplicateState,
    InvalidFunction,
    Registry,
    SaltObject,
    State,
    StateFactory,
)
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase

log = logging.getLogger(__name__)


class MapBuilder:
    def build_map(self, template=None):
        """
        Build from a specific template or just use a default if no template
        is passed to this function.
        """

        map_prefix = textwrap.dedent(
            """\
        #!pyobjects
        from salt.utils.pyobjects import StateFactory
        Service = StateFactory('service')

        {% macro priority(value) %}
            priority = {{ value }}
        {% endmacro %}
        class Samba(Map):
        """
        )

        map_suffix = textwrap.dedent(
            """\
        with Pkg.installed("samba", names=[Samba.server, Samba.client]):
            Service.running("samba", name=Samba.service)
        """
        )

        map_data = {
            "debian": (
                "    class Debian:\n"
                "        server = 'samba'\n"
                "        client = 'samba-client'\n"
                "        service = 'samba'\n"
            ),
            "centos": (
                "    class RougeChapeau:\n"
                "        __match__ = 'RedHat'\n"
                "        server = 'samba'\n"
                "        client = 'samba'\n"
                "        service = 'smb'\n"
            ),
            "ubuntu": (
                "    class Ubuntu:\n"
                "        __grain__ = 'os'\n"
                "        service = 'smbd'\n"
            ),
        }

        if template is None:
            template = textwrap.dedent(
                """\
                {{ ubuntu }}
                {{ centos }}
                {{ debian }}
                """
            )
        full_template = map_prefix + template + map_suffix
        ret = jinja2.Template(full_template).render(**map_data)
        log.debug("built map: \n%s", ret)
        return ret


class StateTests(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.File = StateFactory("file")

    @classmethod
    def tearDownClass(cls):
        cls.File = None

    def setUp(self):
        Registry.empty()
        self.pydmesg_expected = {
            "file.managed": [
                {"group": "root"},
                {"mode": "0755"},
                {"require": [{"file": "/usr/local/bin"}]},
                {"source": "salt://debian/files/pydmesg.py"},
                {"user": "root"},
            ]
        }
        self.pydmesg_salt_expected = OrderedDict(
            [("/usr/local/bin/pydmesg", self.pydmesg_expected)]
        )
        self.pydmesg_kwargs = dict(
            user="root",
            group="root",
            mode="0755",
            source="salt://debian/files/pydmesg.py",
        )

    def tearDown(self):
        self.pydmesg_expected = self.pydmesg_salt_expected = self.pydmesg_kwargs = None

    def test_serialization(self):
        f = State(
            "/usr/local/bin/pydmesg",
            "file",
            "managed",
            require=self.File("/usr/local/bin"),
            **self.pydmesg_kwargs
        )

        self.assertEqual(f(), self.pydmesg_expected)

    def test_factory_serialization(self):
        self.File.managed(
            "/usr/local/bin/pydmesg",
            require=self.File("/usr/local/bin"),
            **self.pydmesg_kwargs
        )

        self.assertEqual(
            Registry.states["/usr/local/bin/pydmesg"], self.pydmesg_expected
        )

    def test_context_manager(self):
        with self.File("/usr/local/bin"):
            pydmesg = self.File.managed("/usr/local/bin/pydmesg", **self.pydmesg_kwargs)

            self.assertEqual(
                Registry.states["/usr/local/bin/pydmesg"], self.pydmesg_expected
            )

            with pydmesg:
                self.File.managed("/tmp/something", owner="root")

                self.assertEqual(
                    Registry.states["/tmp/something"],
                    {
                        "file.managed": [
                            {"owner": "root"},
                            {
                                "require": [
                                    {"file": "/usr/local/bin"},
                                    {"file": "/usr/local/bin/pydmesg"},
                                ]
                            },
                        ]
                    },
                )

    def test_salt_data(self):
        self.File.managed(
            "/usr/local/bin/pydmesg",
            require=self.File("/usr/local/bin"),
            **self.pydmesg_kwargs
        )

        self.assertEqual(
            Registry.states["/usr/local/bin/pydmesg"], self.pydmesg_expected
        )

        self.assertEqual(Registry.salt_data(), self.pydmesg_salt_expected)

        self.assertEqual(Registry.states, OrderedDict())

    def test_duplicates(self):
        def add_dup():
            self.File.managed("dup", name="/dup")

        add_dup()
        self.assertRaises(DuplicateState, add_dup)
        Service = StateFactory("service")
        Service.running("dup", name="dup-service")

        self.assertEqual(
            Registry.states,
            OrderedDict(
                [
                    (
                        "dup",
                        OrderedDict(
                            [
                                ("file.managed", [{"name": "/dup"}]),
                                ("service.running", [{"name": "dup-service"}]),
                            ]
                        ),
                    )
                ]
            ),
        )


class RendererMixin:
    """
    This is a mixin that adds a ``.render()`` method to render a template

    It must come BEFORE ``TestCase`` in the declaration of your test case
    class so that our setUp & tearDown get invoked first, and super can
    trigger the methods in the ``TestCase`` class.
    """

    def setUp(self, *args, **kwargs):
        super().setUp(*args, **kwargs)

        self.root_dir = tempfile.mkdtemp("pyobjects_test_root", dir=RUNTIME_VARS.TMP)
        self.state_tree_dir = os.path.join(self.root_dir, "state_tree")
        self.cache_dir = os.path.join(self.root_dir, "cachedir")
        if not os.path.isdir(self.root_dir):
            os.makedirs(self.root_dir)

        if not os.path.isdir(self.state_tree_dir):
            os.makedirs(self.state_tree_dir)

        if not os.path.isdir(self.cache_dir):
            os.makedirs(self.cache_dir)
        self.config = salt.config.minion_config(None)
        self.config["root_dir"] = self.root_dir
        self.config["state_events"] = False
        self.config["id"] = "match"
        self.config["file_client"] = "local"
        self.config["file_roots"] = dict(base=[self.state_tree_dir])
        self.config["cachedir"] = self.cache_dir
        self.config["test"] = False

    def tearDown(self, *args, **kwargs):
        shutil.rmtree(self.root_dir)
        del self.config
        super().tearDown(*args, **kwargs)

    def write_template_file(self, filename, content):
        full_path = os.path.join(self.state_tree_dir, filename)
        with salt.utils.files.fopen(full_path, "w") as f:
            f.write(content)
        return full_path

    def render(self, template, opts=None, filename=None):
        if opts:
            self.config.update(opts)

        if not filename:
            filename = ".".join([str(uuid.uuid4()), "sls"])
        full_path = self.write_template_file(filename, template)

        state = salt.state.State(self.config)
        return compile_template(
            full_path,
            state.rend,
            state.opts["renderer"],
            state.opts["renderer_blacklist"],
            state.opts["renderer_whitelist"],
        )


class RendererTests(RendererMixin, StateTests, MapBuilder):
    @classmethod
    def setUpClass(cls):
        cls.recursive_map_template = textwrap.dedent(
            """\
        #!pyobjects
        from salt://map.sls import Samba

        class CustomSamba(Samba):
            pass
        """
        )
        cls.recursive_import_template = textwrap.dedent(
            """\
        #!pyobjects
        from salt://recursive_map.sls import CustomSamba

        Pkg.removed("samba-imported", names=[CustomSamba.server, CustomSamba.client])"""
        )
        cls.File = StateFactory("file")

    @classmethod
    def tearDownClass(cls):
        cls.File = None

    @pytest.mark.slow_test
    def test_basic(self):
        basic_template = textwrap.dedent(
            """\
        #!pyobjects
        File.directory('/tmp', mode='1777', owner='root', group='root')
        """
        )
        ret = self.render(basic_template)
        self.assertEqual(
            ret,
            OrderedDict(
                [
                    (
                        "/tmp",
                        {
                            "file.directory": [
                                {"group": "root"},
                                {"mode": "1777"},
                                {"owner": "root"},
                            ]
                        },
                    ),
                ]
            ),
        )
        self.assertEqual(Registry.states, OrderedDict())

    @pytest.mark.slow_test
    def test_invalid_function(self):
        def _test():
            invalid_template = textwrap.dedent(
                """\
            #!pyobjects
            File.fail('/tmp')
            """
            )
            self.render(invalid_template)

        self.assertRaises(InvalidFunction, _test)

    @pytest.mark.slow_test
    def test_include(self):
        include_template = textwrap.dedent(
            """\
        #!pyobjects
        include('http')
        """
        )
        ret = self.render(include_template)
        self.assertEqual(ret, OrderedDict([("include", ["http"])]))

    @pytest.mark.slow_test
    def test_extend(self):
        extend_template = textwrap.dedent(
            """\
        #!pyobjects
        include('http')

        from salt.utils.pyobjects import StateFactory
        Service = StateFactory('service')

        Service.running(extend('apache'), watch=[{'file': '/etc/file'}])
        """
        )
        ret = self.render(
            extend_template, {"grains": {"os_family": "Debian", "os": "Debian"}}
        )
        self.assertEqual(
            ret,
            OrderedDict(
                [
                    ("include", ["http"]),
                    (
                        "extend",
                        OrderedDict(
                            [
                                (
                                    "apache",
                                    {
                                        "service.running": [
                                            {"watch": [{"file": "/etc/file"}]}
                                        ]
                                    },
                                ),
                            ]
                        ),
                    ),
                ]
            ),
        )

    @pytest.mark.slow_test
    def test_sls_imports(self):
        def render_and_assert(template):
            ret = self.render(
                template, {"grains": {"os_family": "Debian", "os": "Debian"}}
            )

            self.assertEqual(
                ret,
                OrderedDict(
                    [
                        (
                            "samba-imported",
                            {"pkg.removed": [{"names": ["samba", "samba-client"]}]},
                        )
                    ]
                ),
            )

        self.write_template_file("map.sls", self.build_map())

        import_template = textwrap.dedent(
            """\
        #!pyobjects
        import salt://map.sls

        Pkg.removed("samba-imported", names=[map.Samba.server, map.Samba.client])
        """
        )
        render_and_assert(import_template)

        from_import_template = textwrap.dedent(
            """\
        #!pyobjects
        # this spacing is like this on purpose to ensure it's stripped properly
        from   salt://map.sls  import     Samba

        Pkg.removed("samba-imported", names=[Samba.server, Samba.client])
        """
        )
        render_and_assert(from_import_template)

        import_as_template = textwrap.dedent(
            """\
        #!pyobjects
        from salt://map.sls import Samba as Other
        Pkg.removed("samba-imported", names=[Other.server, Other.client])
        """
        )
        render_and_assert(import_as_template)

        self.write_template_file("recursive_map.sls", self.recursive_map_template)
        render_and_assert(self.recursive_import_template)

    @pytest.mark.slow_test
    def test_import_scope(self):
        self.write_template_file("map.sls", self.build_map())
        self.write_template_file("recursive_map.sls", self.recursive_map_template)

        def do_render():
            scope_test_import_template = textwrap.dedent(
                """\
            #!pyobjects
            from salt://recursive_map.sls import CustomSamba

            # since we import CustomSamba we should shouldn't be able to see Samba
            Pkg.removed("samba-imported", names=[Samba.server, Samba.client])"""
            )
            ret = self.render(
                scope_test_import_template,
                {"grains": {"os_family": "Debian", "os": "Debian"}},
            )

        self.assertRaises(NameError, do_render)

    @pytest.mark.slow_test
    def test_random_password(self):
        """Test for https://github.com/saltstack/salt/issues/21796"""
        random_password_template = textwrap.dedent(
            """\
        #!pyobjects
        import random, string
        password = ''.join([random.SystemRandom().choice(
                string.ascii_letters + string.digits) for _ in range(20)])
        """
        )
        ret = self.render(random_password_template)

    @pytest.mark.slow_test
    def test_import_random_password(self):
        """Import test for https://github.com/saltstack/salt/issues/21796"""
        random_password_template = textwrap.dedent(
            """\
        #!pyobjects
        import random, string
        password = ''.join([random.SystemRandom().choice(
                string.ascii_letters + string.digits) for _ in range(20)])
        """
        )
        self.write_template_file("password.sls", random_password_template)

        random_password_import_template = textwrap.dedent(
            """\
        #!pyobjects
        from salt://password.sls import password
        """
        )
        ret = self.render(random_password_import_template)

    @pytest.mark.slow_test
    def test_requisite_implicit_list(self):
        """Ensure that the implicit list characteristic works as expected"""
        requisite_implicit_list_template = textwrap.dedent(
            """\
        #!pyobjects
        from salt.utils.pyobjects import StateFactory
        Service = StateFactory('service')

        with Pkg.installed("pkg"):
            Service.running("service", watch=File("file"), require=Cmd("cmd"))
        """
        )
        ret = self.render(
            requisite_implicit_list_template,
            {"grains": {"os_family": "Debian", "os": "Debian"}},
        )

        self.assertEqual(
            ret,
            OrderedDict(
                [
                    ("pkg", OrderedDict([("pkg.installed", [])])),
                    (
                        "service",
                        OrderedDict(
                            [
                                (
                                    "service.running",
                                    [
                                        {"require": [{"cmd": "cmd"}, {"pkg": "pkg"}]},
                                        {"watch": [{"file": "file"}]},
                                    ],
                                )
                            ]
                        ),
                    ),
                ]
            ),
        )


class MapTests(RendererMixin, TestCase, MapBuilder):
    maxDiff = None

    debian_grains = {"os_family": "Debian", "os": "Debian"}
    ubuntu_grains = {"os_family": "Debian", "os": "Ubuntu"}
    centos_grains = {"os_family": "RedHat", "os": "CentOS"}

    debian_attrs = ("samba", "samba-client", "samba")
    ubuntu_attrs = ("samba", "samba-client", "smbd")
    centos_attrs = ("samba", "samba", "smb")

    def samba_with_grains(self, template, grains):
        return self.render(template, {"grains": grains})

    def assert_equal(self, ret, server, client, service):
        self.assertDictEqual(
            ret,
            OrderedDict(
                [
                    (
                        "samba",
                        OrderedDict(
                            [
                                ("pkg.installed", [{"names": [server, client]}]),
                                (
                                    "service.running",
                                    [
                                        {"name": service},
                                        {"require": [{"pkg": "samba"}]},
                                    ],
                                ),
                            ]
                        ),
                    )
                ]
            ),
        )

    def assert_not_equal(self, ret, server, client, service):
        try:
            self.assert_equal(ret, server, client, service)
        except AssertionError:
            pass
        else:
            raise AssertionError("both dicts are equal")

    @pytest.mark.slow_test
    def test_map(self):
        """
        Test declarative ordering
        """
        # With declarative ordering, the ubuntu-specific service name should
        # override the one inherited from debian.
        template = self.build_map(
            textwrap.dedent(
                """\
            {{ debian }}
            {{ centos }}
            {{ ubuntu }}
            """
            )
        )

        ret = self.samba_with_grains(template, self.debian_grains)
        self.assert_equal(ret, *self.debian_attrs)

        ret = self.samba_with_grains(template, self.ubuntu_grains)
        self.assert_equal(ret, *self.ubuntu_attrs)

        ret = self.samba_with_grains(template, self.centos_grains)
        self.assert_equal(ret, *self.centos_attrs)

        # Switching the order, debian should still work fine but ubuntu should
        # no longer match, since the debian service name should override the
        # ubuntu one.
        template = self.build_map(
            textwrap.dedent(
                """\
            {{ ubuntu }}
            {{ debian }}
            """
            )
        )

        ret = self.samba_with_grains(template, self.debian_grains)
        self.assert_equal(ret, *self.debian_attrs)

        ret = self.samba_with_grains(template, self.ubuntu_grains)
        self.assert_not_equal(ret, *self.ubuntu_attrs)

    @pytest.mark.slow_test
    def test_map_with_priority(self):
        """
        With declarative ordering, the debian service name would override the
        ubuntu one since debian comes second. This will test overriding this
        behavior using the priority attribute.
        """
        template = self.build_map(
            textwrap.dedent(
                """\
            {{ priority(('os_family', 'os')) }}
            {{ ubuntu }}
            {{ centos }}
            {{ debian }}
            """
            )
        )

        ret = self.samba_with_grains(template, self.debian_grains)
        self.assert_equal(ret, *self.debian_attrs)

        ret = self.samba_with_grains(template, self.ubuntu_grains)
        self.assert_equal(ret, *self.ubuntu_attrs)

        ret = self.samba_with_grains(template, self.centos_grains)
        self.assert_equal(ret, *self.centos_attrs)


class SaltObjectTests(TestCase):
    def test_salt_object(self):
        def attr_fail():
            Salt.fail.blah()

        def times2(x):
            return x * 2

        __salt__ = {"math.times2": times2}

        Salt = SaltObject(__salt__)

        self.assertRaises(AttributeError, attr_fail)
        self.assertEqual(Salt.math.times2, times2)
        self.assertEqual(Salt.math.times2(2), 4)



"""
An external pillar module for getting credentials from confidant.

Configuring the Confidant module
================================

The module can be configured via ext_pillar in the minion config:

.. code-block:: yaml

ext_pillar:
  - confidant:
      profile:
        # The URL of the confidant web service
        url: 'https://confidant-production.example.com'
        # The context to use for KMS authentication
        auth_context:
        from: example-production-iad
        to: confidant-production-iad
        user_type: service
        # The KMS master key to use for authentication
        auth_key: "alias/authnz"
        # Cache file for KMS auth token
        token_cache_file: /run/confidant/confidant_token
        # The duration of the validity of a token, in minutes
        token_duration: 60
        # key, keyid and region can be defined in the profile, but it's
        # generally best to use IAM roles or environment variables for AWS
        # auth.
        keyid: 98nh9h9h908h09kjjk
        key: jhf908gyeghehe0he0g8h9u0j0n0n09hj09h0
        region: us-east-1

:depends: confidant-common, confidant-client

Module Documentation
====================
"""

import copy
import logging

try:
    # pylint: disable=no-name-in-module
    import confidant.client
    import confidant.formatter

    HAS_LIBS = True
    # pylint: enable=no-name-in-module
except ImportError:
    HAS_LIBS = False

# Set up logging
log = logging.getLogger(__name__)

__virtualname__ = "confidant"


def __virtual__():
    """
    Only return if requests and boto are installed.
    """
    if HAS_LIBS:
        return __virtualname__
    else:
        return False


def ext_pillar(minion_id, pillar, profile=None):
    """
    Read pillar data from Confidant via its API.
    """
    if profile is None:
        profile = {}
    # default to returning failure
    ret = {
        "credentials_result": False,
        "credentials": None,
        "credentials_metadata": None,
    }
    profile_data = copy.deepcopy(profile)
    if profile_data.get("disabled", False):
        ret["result"] = True
        return ret
    token_version = profile_data.get("token_version", 1)
    try:
        url = profile_data["url"]
        auth_key = profile_data["auth_key"]
        auth_context = profile_data["auth_context"]
        role = auth_context["from"]
    except (KeyError, TypeError):
        msg = "profile has undefined url, auth_key or auth_context"
        log.debug(msg)
        return ret
    region = profile_data.get("region", "us-east-1")
    token_duration = profile_data.get("token_duration", 60)
    retries = profile_data.get("retries", 5)
    token_cache_file = profile_data.get("token_cache_file")
    backoff = profile_data.get("backoff", 1)
    client = confidant.client.ConfidantClient(
        url,
        auth_key,
        auth_context,
        token_lifetime=token_duration,
        token_version=token_version,
        token_cache_file=token_cache_file,
        region=region,
        retries=retries,
        backoff=backoff,
    )
    try:
        data = client.get_service(role, decrypt_blind=True)
    except confidant.client.TokenCreationError:
        return ret
    if not data["result"]:
        return ret
    ret = confidant.formatter.combined_credential_pair_format(data)
    ret["credentials_result"] = True
    return ret



"""
Copy pillar data from a bucket in Amazon S3

The S3 pillar can be configured in the master config file with the following
options

.. code-block:: yaml

    ext_pillar:
      - s3:
          bucket: my.fancy.pillar.bucket
          keyid: KASKFJWAKJASJKDAJKSD
          key: ksladfDLKDALSFKSD93q032sdDasdfasdflsadkf
          multiple_env: False
          environment: base
          prefix: somewhere/overthere
          verify_ssl: True
          service_url: s3.amazonaws.com
          kms_keyid: 01234567-89ab-cdef-0123-4567890abcde
          s3_cache_expire: 30
          s3_sync_on_update: True
          path_style: False
          https_enable: True

The ``bucket`` parameter specifies the target S3 bucket. It is required.

The ``keyid`` parameter specifies the key id to use when access the S3 bucket.
If it is not provided, an attempt to fetch it from EC2 instance meta-data will
be made.

The ``key`` parameter specifies the key to use when access the S3 bucket. If it
is not provided, an attempt to fetch it from EC2 instance meta-data will be made.

The ``multiple_env`` defaults to False. It specifies whether the pillar should
interpret top level folders as pillar environments (see mode section below).

The ``environment`` defaults to 'base'. It specifies which environment the
bucket represents when in single environments mode (see mode section below). It
is ignored if multiple_env is True.

The ``prefix`` defaults to ''. It specifies a key prefix to use when searching
for data in the bucket for the pillar. It works when multiple_env is True or False.
Essentially it tells ext_pillar to look for your pillar data in a 'subdirectory'
of your S3 bucket

The ``verify_ssl`` parameter defaults to True. It specifies whether to check for
valid S3 SSL certificates. *NOTE* If you use bucket names with periods, this
must be set to False else an invalid certificate error will be thrown (issue
#12200).

The ``service_url`` parameter defaults to 's3.amazonaws.com'. It specifies the
base url to use for accessing S3.

The ``kms_keyid`` parameter is optional. It specifies the ID of the Key
Management Service (KMS) master key that was used to encrypt the object.

The ``s3_cache_expire`` parameter defaults to 30s. It specifies expiration
time of S3 metadata cache file.

The ``s3_sync_on_update`` parameter defaults to True. It specifies if cache
is synced on update rather than jit.

The ``path_style`` parameter defaults to False. It specifies whether to use
path style requests or dns style requests

The ``https_enable`` parameter defaults to True. It specifies whether to use
https protocol or http protocol

This pillar can operate in two modes, single environment per bucket or multiple
environments per bucket.

Single environment mode must have this bucket structure:

.. code-block:: text

    s3://<bucket name>/<prefix>/<files>

Multiple environment mode must have this bucket structure:

.. code-block:: text

    s3://<bucket name>/<prefix>/<environment>/<files>

If you wish to define your pillar data entirely within S3 it's recommended
that you use the `prefix=` parameter and specify one entry in ext_pillar
for each environment rather than specifying multiple_env. This is due
to issue #22471 (https://github.com/saltstack/salt/issues/22471)
"""


import logging
import os
import pickle
import time
import urllib.parse
from copy import deepcopy

import salt.utils.files
import salt.utils.hashutils
from salt.pillar import Pillar

log = logging.getLogger(__name__)


class S3Credentials:
    def __init__(
        self,
        key,
        keyid,
        bucket,
        service_url,
        verify_ssl=True,
        kms_keyid=None,
        location=None,
        path_style=False,
        https_enable=True,
    ):
        self.key = key
        self.keyid = keyid
        self.kms_keyid = kms_keyid
        self.bucket = bucket
        self.service_url = service_url
        self.verify_ssl = verify_ssl
        self.location = location
        self.path_style = path_style
        self.https_enable = https_enable


def ext_pillar(
    minion_id,
    pillar,  # pylint: disable=W0613
    bucket,
    key=None,
    keyid=None,
    verify_ssl=True,
    location=None,
    multiple_env=False,
    environment="base",
    prefix="",
    service_url=None,
    kms_keyid=None,
    s3_cache_expire=30,  # cache for 30 seconds
    s3_sync_on_update=True,  # sync cache on update rather than jit
    path_style=False,
    https_enable=True,
):

    """
    Execute a command and read the output as YAML
    """

    s3_creds = S3Credentials(
        key,
        keyid,
        bucket,
        service_url,
        verify_ssl,
        kms_keyid,
        location,
        path_style,
        https_enable,
    )

    # normpath is needed to remove appended '/' if root is empty string.
    pillar_dir = os.path.normpath(os.path.join(_get_cache_dir(), environment, bucket))
    if prefix:
        pillar_dir = os.path.normpath(os.path.join(pillar_dir, prefix))

    if __opts__["pillar_roots"].get(environment, []) == [pillar_dir]:
        return {}

    metadata = _init(
        s3_creds, bucket, multiple_env, environment, prefix, s3_cache_expire
    )

    if s3_sync_on_update:
        # sync the buckets to the local cache
        log.info("Syncing local pillar cache from S3...")
        for saltenv, env_meta in metadata.items():
            for bucket, files in _find_files(env_meta).items():
                for file_path in files:
                    cached_file_path = _get_cached_file_name(bucket, saltenv, file_path)
                    log.info("%s - %s : %s", bucket, saltenv, file_path)
                    # load the file from S3 if not in the cache or too old
                    _get_file_from_s3(
                        s3_creds, metadata, saltenv, bucket, file_path, cached_file_path
                    )

        log.info("Sync local pillar cache from S3 completed.")

    opts = deepcopy(__opts__)
    opts["pillar_roots"][environment] = (
        [os.path.join(pillar_dir, environment)] if multiple_env else [pillar_dir]
    )

    # Avoid recursively re-adding this same pillar
    opts["ext_pillar"] = [x for x in opts["ext_pillar"] if "s3" not in x]

    pil = Pillar(opts, __grains__, minion_id, environment)

    compiled_pillar = pil.compile_pillar(ext=False)

    return compiled_pillar


def _init(creds, bucket, multiple_env, environment, prefix, s3_cache_expire):
    """
    Connect to S3 and download the metadata for each file in all buckets
    specified and cache the data to disk.
    """

    cache_file = _get_buckets_cache_filename(bucket, prefix)
    exp = time.time() - s3_cache_expire

    # check if cache_file exists and its mtime
    if os.path.isfile(cache_file):
        cache_file_mtime = os.path.getmtime(cache_file)
    else:
        # file does not exists then set mtime to 0 (aka epoch)
        cache_file_mtime = 0

    expired = cache_file_mtime <= exp

    log.debug(
        "S3 bucket cache file %s is %sexpired, mtime_diff=%ss, expiration=%ss",
        cache_file,
        "" if expired else "not ",
        cache_file_mtime - exp,
        s3_cache_expire,
    )

    if expired:
        pillars = _refresh_buckets_cache_file(
            creds, cache_file, multiple_env, environment, prefix
        )
    else:
        pillars = _read_buckets_cache_file(cache_file)

    log.debug("S3 bucket retrieved pillars %s", pillars)
    return pillars


def _get_cache_dir():
    """
    Get pillar cache directory. Initialize it if it does not exist.
    """

    cache_dir = os.path.join(__opts__["cachedir"], "pillar_s3fs")

    if not os.path.isdir(cache_dir):
        log.debug("Initializing S3 Pillar Cache")
        os.makedirs(cache_dir)

    return cache_dir


def _get_cached_file_name(bucket, saltenv, path):
    """
    Return the cached file name for a bucket path file
    """

    file_path = os.path.join(_get_cache_dir(), saltenv, bucket, path)

    # make sure bucket and saltenv directories exist
    if not os.path.exists(os.path.dirname(file_path)):
        os.makedirs(os.path.dirname(file_path))

    return file_path


def _get_buckets_cache_filename(bucket, prefix):
    """
    Return the filename of the cache for bucket contents.
    Create the path if it does not exist.
    """

    cache_dir = _get_cache_dir()
    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir)

    return os.path.join(cache_dir, "{}-{}-files.cache".format(bucket, prefix))


def _refresh_buckets_cache_file(creds, cache_file, multiple_env, environment, prefix):
    """
    Retrieve the content of all buckets and cache the metadata to the buckets
    cache file
    """

    # helper s3 query function
    def __get_s3_meta(continuation_token=None):
        # We want to use ListObjectsV2 so we get the NextContinuationToken
        params = {"prefix": prefix, "list-type": 2}

        if continuation_token:
            params["continuation-token"] = continuation_token

        return __utils__["s3.query"](
            key=creds.key,
            keyid=creds.keyid,
            kms_keyid=creds.kms_keyid,
            bucket=creds.bucket,
            service_url=creds.service_url,
            verify_ssl=creds.verify_ssl,
            location=creds.location,
            return_bin=False,
            params=params,
            path_style=creds.path_style,
            https_enable=creds.https_enable,
        )

    # grab only the files/dirs in the bucket
    def __get_pillar_files_from_s3_meta(s3_meta):
        return [k for k in s3_meta if "Key" in k]

    # pull out the environment dirs (e.g. the root dirs)
    def __get_pillar_environments(files):
        environments = [(os.path.dirname(k["Key"]).split("/", 1))[0] for k in files]
        return set(environments)

    def __get_continuation_token(s3_meta):
        return next(
            (
                item.get("NextContinuationToken")
                for item in s3_meta
                if item.get("NextContinuationToken")
            ),
            None,
        )

    log.debug("Refreshing S3 buckets pillar cache file")

    metadata = {}
    bucket = creds.bucket

    if not multiple_env:
        # Single environment per bucket
        log.debug("Single environment per bucket mode")

        bucket_files = {}
        s3_meta = __get_s3_meta()

        # s3 query returned something
        if s3_meta:
            bucket_files[bucket] = __get_pillar_files_from_s3_meta(s3_meta)

            # Check if we have a NextContinuationToken and loop until we don't
            while True:
                continuation_token = __get_continuation_token(s3_meta)
                if not continuation_token:
                    break
                s3_meta = __get_s3_meta(continuation_token)
                bucket_files[bucket] += __get_pillar_files_from_s3_meta(s3_meta)

            metadata[environment] = bucket_files

    else:
        # Multiple environments per buckets
        log.debug("Multiple environment per bucket mode")
        s3_meta = __get_s3_meta()

        # s3 query returned data
        if s3_meta:
            files = __get_pillar_files_from_s3_meta(s3_meta)

            # Check if we have a NextContinuationToken and loop until we don't
            while True:
                continuation_token = __get_continuation_token(s3_meta)
                if not continuation_token:
                    break
                s3_meta = __get_s3_meta(continuation_token)
                files += __get_pillar_files_from_s3_meta(s3_meta)

            environments = __get_pillar_environments(files)

            # pull out the files for the environment
            for saltenv in environments:
                # grab only files/dirs that match this saltenv.
                env_files = [k for k in files if k["Key"].startswith(saltenv)]

                if saltenv not in metadata:
                    metadata[saltenv] = {}

                if bucket not in metadata[saltenv]:
                    metadata[saltenv][bucket] = []

                metadata[saltenv][bucket] += env_files

    # write the metadata to disk
    if os.path.isfile(cache_file):
        os.remove(cache_file)

    log.debug("Writing S3 buckets pillar cache file")

    with salt.utils.files.fopen(cache_file, "wb") as fp_:
        pickle.dump(metadata, fp_)

    return metadata


def _read_buckets_cache_file(cache_file):
    """
    Return the contents of the buckets cache file
    """

    log.debug("Reading buckets cache file")

    with salt.utils.files.fopen(cache_file, "rb") as fp_:
        data = pickle.load(fp_)

    return data


def _find_files(metadata):
    """
    Looks for all the files in the S3 bucket cache metadata
    """

    ret = {}

    for bucket, data in metadata.items():
        if bucket not in ret:
            ret[bucket] = []

        # grab the paths from the metadata
        filePaths = [k["Key"] for k in data]
        # filter out the dirs
        ret[bucket] += [k for k in filePaths if not k.endswith("/")]

    return ret


def _find_file_meta(metadata, bucket, saltenv, path):
    """
    Looks for a file's metadata in the S3 bucket cache file
    """

    env_meta = metadata[saltenv] if saltenv in metadata else {}
    bucket_meta = env_meta[bucket] if bucket in env_meta else {}
    files_meta = list(list(filter((lambda k: "Key" in k), bucket_meta)))

    for item_meta in files_meta:
        if "Key" in item_meta and item_meta["Key"] == path:
            return item_meta


def _get_file_from_s3(creds, metadata, saltenv, bucket, path, cached_file_path):
    """
    Checks the local cache for the file, if it's old or missing go grab the
    file from S3 and update the cache
    """

    # check the local cache...
    if os.path.isfile(cached_file_path):
        file_meta = _find_file_meta(metadata, bucket, saltenv, path)
        file_md5 = (
            "".join(list(filter(str.isalnum, file_meta["ETag"]))) if file_meta else None
        )

        cached_md5 = salt.utils.hashutils.get_hash(cached_file_path, "md5")

        # hashes match we have a cache hit
        log.debug(
            "Cached file: path=%s, md5=%s, etag=%s",
            cached_file_path,
            cached_md5,
            file_md5,
        )
        if cached_md5 == file_md5:
            return

    # ... or get the file from S3
    __utils__["s3.query"](
        key=creds.key,
        keyid=creds.keyid,
        kms_keyid=creds.kms_keyid,
        bucket=bucket,
        service_url=creds.service_url,
        path=urllib.parse.quote(path),
        local_file=cached_file_path,
        verify_ssl=creds.verify_ssl,
        location=creds.location,
        path_style=creds.path_style,
        https_enable=creds.https_enable,
    )



"""
    salt._logging.handlers
    ~~~~~~~~~~~~~~~~~~~~~~

    Salt's logging handlers
"""

import copy
import logging
import logging.handlers
import queue as _queue
import sys
from collections import deque

from salt._logging.mixins import ExcInfoOnLogLevelFormatMixin
from salt.utils.versions import warn_until_date

log = logging.getLogger(__name__)


class TemporaryLoggingHandler(logging.NullHandler):
    """
    This logging handler will store all the log records up to its maximum
    queue size at which stage the first messages stored will be dropped.

    Should only be used as a temporary logging handler, while the logging
    system is not fully configured.

    Once configured, pass any logging handlers that should have received the
    initial log messages to the function
    :func:`TemporaryLoggingHandler.sync_with_handlers` and all stored log
    records will be dispatched to the provided handlers.

    .. versionadded:: 0.17.0
    """

    def __init__(self, level=logging.NOTSET, max_queue_size=10000):
        warn_until_date(
            "20240101",
            "Please stop using '{name}.TemporaryLoggingHandler'. "
            "'{name}.TemporaryLoggingHandler' will go away after "
            "{{date}}.".format(name=__name__),
        )
        super().__init__(level=level)
        self.__messages = deque(maxlen=max_queue_size)

    def handle(self, record):
        self.acquire()
        self.__messages.append(record)
        self.release()

    def sync_with_handlers(self, handlers=()):
        """
        Sync the stored log records to the provided log handlers.
        """
        if not handlers:
            return

        while self.__messages:
            record = self.__messages.popleft()
            for handler in handlers:
                if handler.level > record.levelno:
                    # If the handler's level is higher than the log record one,
                    # it should not handle the log record
                    continue
                handler.handle(record)


class StreamHandler(ExcInfoOnLogLevelFormatMixin, logging.StreamHandler):
    """
    Stream handler which properly handles exc_info on a per handler basis
    """


class DeferredStreamHandler(StreamHandler):
    """
    This logging handler will store all the log records up to its maximum
    queue size at which stage the first messages stored will be dropped.

    Should only be used as a temporary logging handler, while the logging
    system is not fully configured.

    Once configured, pass any logging handlers that should have received the
    initial log messages to the function
    :func:`DeferredStreamHandler.sync_with_handlers` and all stored log
    records will be dispatched to the provided handlers.

    If anything goes wrong before logging is properly setup, all stored messages
    will be flushed to the handler's stream, ie, written to console.

    .. versionadded:: 3005
    """

    def __init__(self, stream, max_queue_size=10000):
        super().__init__(stream)
        self.__messages = deque(maxlen=max_queue_size)
        self.__emitting = False

    def handle(self, record):
        self.acquire()
        self.__messages.append(record)
        self.release()

    def flush(self):
        if self.__emitting:
            # We set the flushing flag because the stream handler actually calls flush when
            # emitting a log record and we don't want to cause a RecursionError
            return
        while self.__messages:
            try:
                self.__emitting = True
                record = self.__messages.popleft()
                # We call the parent's class handle method so it's actually
                # handled and not queued back
                # However, temporarily
                super().handle(record)
            finally:
                self.__emitting = False
        super().flush()

    def sync_with_handlers(self, handlers=()):
        """
        Sync the stored log records to the provided log handlers.
        """
        while self.__messages:
            record = self.__messages.popleft()
            for handler in handlers:
                if handler is self:
                    continue
                handler.handle(record)


class FileHandler(ExcInfoOnLogLevelFormatMixin, logging.FileHandler):
    """
    File handler which properly handles exc_info on a per handler basis
    """


class SysLogHandler(ExcInfoOnLogLevelFormatMixin, logging.handlers.SysLogHandler):
    """
    Syslog handler which properly handles exc_info on a per handler basis
    """

    def handleError(self, record):
        """
        Override the default error handling mechanism for py3
        Deal with syslog os errors when the log file does not exist
        """
        handled = False
        if sys.stderr and sys.version_info >= (3, 5, 4):
            exc_type, exc, exc_traceback = sys.exc_info()
            try:
                if exc_type.__name__ in "FileNotFoundError":
                    sys.stderr.write(
                        "[WARNING ] The log_file does not exist. Logging not "
                        "setup correctly or syslog service not started.\n"
                    )
                    handled = True
            finally:
                # 'del' recommended. See documentation of
                # 'sys.exc_info()' for details.
                del exc_type, exc, exc_traceback

        if not handled:
            super().handleError(record)


class RotatingFileHandler(
    ExcInfoOnLogLevelFormatMixin, logging.handlers.RotatingFileHandler
):
    """
    Rotating file handler which properly handles exc_info on a per handler basis
    """

    def handleError(self, record):
        """
        Override the default error handling mechanism

        Deal with log file rotation errors due to log file in use
        more softly.
        """
        handled = False

        # Can't use "salt.utils.platform.is_windows()" in this file
        if (
            sys.platform.startswith("win") and logging.raiseExceptions and sys.stderr
        ):  # see Python issue 13807
            exc_type, exc, exc_traceback = sys.exc_info()
            try:
                # PermissionError is used since Python 3.3.
                # OSError is used for previous versions of Python.
                if (
                    exc_type.__name__ in ("PermissionError", "OSError")
                    and exc.winerror == 32
                ):
                    if self.level <= logging.WARNING:
                        sys.stderr.write(
                            '[WARNING ] Unable to rotate the log file "{}" '
                            "because it is in use\n".format(self.baseFilename)
                        )
                    handled = True
            finally:
                # 'del' recommended. See documentation of
                # 'sys.exc_info()' for details.
                del exc_type, exc, exc_traceback

        if not handled:
            super().handleError(record)


class WatchedFileHandler(
    ExcInfoOnLogLevelFormatMixin, logging.handlers.WatchedFileHandler
):
    """
    Watched file handler which properly handles exc_info on a per handler basis
    """


if sys.version_info < (3, 7):
    # On python versions lower than 3.7, we sill subclass and overwrite prepare to include the fix for:
    #  https://bugs.python.org/issue35726
    class QueueHandler(
        ExcInfoOnLogLevelFormatMixin, logging.handlers.QueueHandler
    ):  # pylint: disable=no-member,inconsistent-mro
        def __init__(self, queue):  # pylint: disable=useless-super-delegation
            super().__init__(queue)
            warn_until_date(
                "20240101",
                "Please stop using '{name}.QueueHandler' and instead "
                "use 'logging.handlers.QueueHandler'. "
                "'{name}.QueueHandler' will go away after "
                "{{date}}.".format(name=__name__),
            )

        def enqueue(self, record):
            """
            Enqueue a record.

            The base implementation uses put_nowait. You may want to override
            this method if you want to use blocking, timeouts or custom queue
            implementations.
            """
            try:
                self.queue.put_nowait(record)
            except _queue.Full:
                sys.stderr.write(
                    "[WARNING ] Message queue is full, "
                    'unable to write "{}" to log.\n'.format(record)
                )

        def prepare(self, record):
            """
            Prepares a record for queuing. The object returned by this method is
            enqueued.
            The base implementation formats the record to merge the message
            and arguments, and removes unpickleable items from the record
            in-place.
            You might want to override this method if you want to convert
            the record to a dict or JSON string, or send a modified copy
            of the record while leaving the original intact.
            """
            # The format operation gets traceback text into record.exc_text
            # (if there's exception data), and also returns the formatted
            # message. We can then use this to replace the original
            # msg + args, as these might be unpickleable. We also zap the
            # exc_info and exc_text attributes, as they are no longer
            # needed and, if not None, will typically not be pickleable.
            msg = self.format(record)
            # bpo-35726: make copy of record to avoid affecting other handlers in the chain.
            record = copy.copy(record)
            record.message = msg
            record.msg = msg
            record.args = None
            record.exc_info = None
            record.exc_text = None
            return record

else:

    class QueueHandler(
        ExcInfoOnLogLevelFormatMixin, logging.handlers.QueueHandler
    ):  # pylint: disable=no-member,inconsistent-mro
        def __init__(self, queue):  # pylint: disable=useless-super-delegation
            super().__init__(queue)
            warn_until_date(
                "20240101",
                "Please stop using '{name}.QueueHandler' and instead "
                "use 'logging.handlers.QueueHandler'. "
                "'{name}.QueueHandler' will go away after "
                "{{date}}.".format(name=__name__),
            )

        def enqueue(self, record):
            """
            Enqueue a record.

            The base implementation uses put_nowait. You may want to override
            this method if you want to use blocking, timeouts or custom queue
            implementations.
            """
            try:
                self.queue.put_nowait(record)
            except _queue.Full:
                sys.stderr.write(
                    "[WARNING ] Message queue is full, "
                    'unable to write "{}" to log.\n'.format(record)
                )



"""
Starting or restarting of services and daemons
==============================================

Services are defined as system daemons and are typically launched using system
init or rc scripts. This service state uses whichever service module is loaded
on the minion with the virtualname of ``service``. Services can be defined as
either running or dead.

If you need to know if your init system is supported, see the list of supported
:mod:`service modules <salt.modules.service.py>` for your desired init system
(systemd, sysvinit, launchctl, etc.).

Note that Salt's service execution module, and therefore this service state,
uses OS grains to ascertain which service module should be loaded and used to
execute service functions. As existing distributions change init systems or
new distributions are created, OS detection can sometimes be incomplete.
If your service states are running into trouble with init system detection,
please see the :ref:`Overriding Virtual Module Providers <module-provider-override>`
section of Salt's module documentation to work around possible errors.

For services managed by systemd, the systemd_service module includes a built-in
feature to reload the daemon when unit files are changed or extended. This
feature is used automatically by the service state and the systemd_service
module when running on a systemd minion, so there is no need to set up your own
methods of reloading the daemon. If you need to manually reload the daemon for
some reason, you can use the :func:`systemd_service.systemctl_reload <salt.modules.systemd_service.systemctl_reload>` function provided by Salt.

.. note::
    The current status of a service is determined by the return code of the init/rc
    script status command. A status return code of 0 it is considered running.  Any
    other return code is considered dead.

.. code-block:: yaml

    httpd:
      service.running: []

The service can also be set to start at runtime via the enable option:

.. code-block:: yaml

    openvpn:
      service.running:
        - enable: True

By default if a service is triggered to refresh due to a watch statement the
service is restarted. If the desired behavior is to reload the service, then
set the reload value to True:

.. code-block:: yaml

    redis:
      service.running:
        - enable: True
        - reload: True
        - watch:
          - pkg: redis

.. note::

    More details regarding ``watch`` can be found in the
    :ref:`Requisites <requisites>` documentation.

"""

import logging
import time

import salt.utils.data
import salt.utils.platform
from salt.exceptions import CommandExecutionError
from salt.utils.args import get_function_argspec as _argspec
from salt.utils.systemd import booted

SYSTEMD_ONLY = ("no_block", "unmask", "unmask_runtime")

log = logging.getLogger(__name__)

__virtualname__ = "service"


def __virtual__():
    """
    Only make these states available if a service provider has been detected or
    assigned for this minion
    """
    if "service.start" in __salt__:
        return __virtualname__
    else:
        return (
            False,
            "No service execution module loaded: "
            "check support for service management on {} "
            "".format(__grains__.get("osfinger", __grains__["os"])),
        )


# Double-asterisk deliberately not used here
def _get_systemd_only(func, kwargs):
    if not hasattr(_get_systemd_only, "HAS_SYSTEMD"):
        setattr(_get_systemd_only, "HAS_SYSTEMD", booted())

    ret = {}
    warnings = []
    valid_args = _argspec(func).args
    for systemd_arg in SYSTEMD_ONLY:
        if systemd_arg in kwargs and systemd_arg in valid_args:
            if _get_systemd_only.HAS_SYSTEMD:
                ret[systemd_arg] = kwargs[systemd_arg]
            else:
                warnings.append(
                    "The '{}' argument is not supported by this platform".format(
                        systemd_arg
                    )
                )
    return ret, warnings


def _add_warnings(ret, warnings):
    current_warnings = ret.setdefault("warnings", [])
    current_warnings.extend([x for x in warnings if x not in current_warnings])


def _enabled_used_error(ret):
    """
    Warn of potential typo.
    """
    ret["result"] = False
    ret["comment"] = (
        'Service {} uses non-existent option "enabled".  '
        'Perhaps "enable" option was intended?'.format(ret["name"])
    )
    return ret


def _enable(name, started, result=True, **kwargs):
    """
    Enable the service
    """
    ret = {}

    # is service available?
    try:
        if not _available(name, ret):
            return ret
    except CommandExecutionError as exc:
        ret["result"] = False
        ret["comment"] = exc.strerror
        return ret

    # Set default expected result
    ret["result"] = result

    # Check to see if this minion supports enable
    if "service.enable" not in __salt__ or "service.enabled" not in __salt__:
        if started is True:
            ret[
                "comment"
            ] = "Enable is not available on this minion, service {} started".format(
                name
            )
        elif started is None:
            ret["comment"] = (
                "Enable is not available on this minion,"
                " service {} is in the desired state".format(name)
            )
        else:
            ret[
                "comment"
            ] = "Enable is not available on this minion, service {} is dead".format(
                name
            )
        return ret

    # Service can be enabled
    before_toggle_enable_status = __salt__["service.enabled"](name, **kwargs)
    if before_toggle_enable_status:
        # Service is enabled
        if started is True:
            ret["comment"] = "Service {} is already enabled, and is running".format(
                name
            )
        elif started is None:
            # always be sure in this case to reset the changes dict
            ret["changes"] = {}
            ret[
                "comment"
            ] = "Service {} is already enabled, and is in the desired state".format(
                name
            )
        else:
            ret["comment"] = "Service {} is already enabled, and is dead".format(name)
        return ret

    # Service needs to be enabled
    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "Service {} set to be enabled".format(name)
        return ret

    try:
        if __salt__["service.enable"](name, **kwargs):
            # Service has been enabled
            ret["changes"] = {}
            after_toggle_enable_status = __salt__["service.enabled"](name, **kwargs)
            # on upstart, certain services like apparmor will always return
            # False, even if correctly activated
            # do not trigger a change
            if before_toggle_enable_status != after_toggle_enable_status:
                ret["changes"][name] = True
            if started is True:
                ret["comment"] = "Service {} has been enabled, and is running".format(
                    name
                )
            elif started is None:
                ret[
                    "comment"
                ] = "Service {} has been enabled, and is in the desired state".format(
                    name
                )
            else:
                ret["comment"] = "Service {} has been enabled, and is dead".format(name)
            return ret
    except CommandExecutionError as exc:
        enable_error = exc.strerror
    else:
        enable_error = False

    # Service failed to be enabled
    ret["result"] = False
    if started is True:
        ret["comment"] = (
            "Failed when setting service {} to start at boot,"
            " but the service is running".format(name)
        )
    elif started is None:
        ret["comment"] = (
            "Failed when setting service {} to start at boot,"
            " but the service was already running".format(name)
        )
    else:
        ret[
            "comment"
        ] = "Failed when setting service {} to start at boot, and the service is dead".format(
            name
        )

    if enable_error:
        ret["comment"] += ". Additional information follows:\n\n{}".format(enable_error)

    return ret


def _disable(name, started, result=True, **kwargs):
    """
    Disable the service
    """
    ret = {}

    # is service available?
    try:
        if not _available(name, ret):
            ret["result"] = True
            return ret
    except CommandExecutionError as exc:
        ret["result"] = False
        ret["comment"] = exc.strerror
        return ret

    # Set default expected result
    ret["result"] = result

    # is enable/disable available?
    if "service.disable" not in __salt__ or "service.disabled" not in __salt__:
        if started is True:
            ret[
                "comment"
            ] = "Disable is not available on this minion, service {} started".format(
                name
            )
        elif started is None:
            ret["comment"] = (
                "Disable is not available on this minion,"
                " service {} is in the desired state".format(name)
            )
        else:
            ret[
                "comment"
            ] = "Disable is not available on this minion, service {} is dead".format(
                name
            )
        return ret

    # Service can be disabled
    if salt.utils.platform.is_windows():
        # service.disabled in Windows returns True for services that are set to
        # Manual start, so we need to check specifically for Disabled
        before_toggle_disable_status = __salt__["service.info"](name)["StartType"] in [
            "Disabled"
        ]
    else:
        before_toggle_disable_status = __salt__["service.disabled"](name)
    if before_toggle_disable_status:
        # Service is disabled
        if started is True:
            ret["comment"] = "Service {} is already disabled, and is running".format(
                name
            )
        elif started is None:
            # always be sure in this case to reset the changes dict
            ret["changes"] = {}
            ret[
                "comment"
            ] = "Service {} is already disabled, and is in the desired state".format(
                name
            )
        else:
            ret["comment"] = "Service {} is already disabled, and is dead".format(name)
        return ret

    # Service needs to be disabled
    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "Service {} set to be disabled".format(name)
        return ret

    if __salt__["service.disable"](name, **kwargs):
        # Service has been disabled
        ret["changes"] = {}
        after_toggle_disable_status = __salt__["service.disabled"](name)
        # on upstart, certain services like apparmor will always return
        # False, even if correctly activated
        # do not trigger a change
        if before_toggle_disable_status != after_toggle_disable_status:
            ret["changes"][name] = True
        if started is True:
            ret["comment"] = "Service {} has been disabled, and is running".format(name)
        elif started is None:
            ret[
                "comment"
            ] = "Service {} has been disabled, and is in the desired state".format(name)
        else:
            ret["comment"] = "Service {} has been disabled, and is dead".format(name)
        return ret

    # Service failed to be disabled
    ret["result"] = False
    if started is True:
        ret[
            "comment"
        ] = "Failed when setting service {} to not start at boot, and is running".format(
            name
        )
    elif started is None:
        ret["comment"] = (
            "Failed when setting service {} to not start"
            " at boot, but the service was already running".format(name)
        )
    else:
        ret["comment"] = (
            "Failed when setting service {} to not start"
            " at boot, and the service is dead".format(name)
        )
    return ret


def _offline():
    return "service.offline" in __salt__ and __salt__["service.offline"]()


def _available(name, ret):
    """
    Check if the service is available
    """
    avail = False
    if "service.available" in __salt__:
        avail = __salt__["service.available"](name)
    elif "service.get_all" in __salt__:
        avail = name in __salt__["service.get_all"]()
    if not avail:
        ret["result"] = False
        ret["comment"] = "The named service {} is not available".format(name)
    return avail


def running(name, enable=None, sig=None, init_delay=None, **kwargs):
    """
    Ensure that the service is running

    name
        The name of the init or rc script used to manage the service

    enable
        Set the service to be enabled at boot time, ``True`` sets the service
        to be enabled, ``False`` sets the named service to be disabled. The
        default is ``None``, which does not enable or disable anything.

    sig
        The string to search for when looking for the service process with ps

    init_delay
        Some services may not be truly available for a short period after their
        startup script indicates to the system that they are. Provide an
        'init_delay' to specify that this state should wait an additional given
        number of seconds after a service has started before returning. Useful
        for requisite states wherein a dependent state might assume a service
        has started but is not yet fully initialized.

    no_block : False
        **For systemd minions only.** Starts the service using ``--no-block``.

        .. versionadded:: 2017.7.0

    timeout
        **For Windows minions only.**

        The time in seconds to wait for the service to start before returning.
        Default is the default for :py:func:`win_service.start
        <salt.modules.win_service.start>`.

        .. versionadded:: 2017.7.9,2018.3.4

    unmask : False
        **For systemd minions only.** Set to ``True`` to remove an indefinite
        mask before attempting to start the service.

        .. versionadded:: 2017.7.0
            In previous releases, Salt would simply unmask a service before
            making any changes. This behavior is no longer the default.

    unmask_runtime : False
        **For systemd minions only.** Set to ``True`` to remove a runtime mask
        before attempting to start the service.

        .. versionadded:: 2017.7.0
            In previous releases, Salt would simply unmask a service before
            making any changes. This behavior is no longer the default.

    wait : 3
        **For systemd minions only.** Passed through when using
        :py:func:`service.status <salt.modules.systemd_service.status>` to
        determine whether the service is running or not.

        .. versionadded:: 2019.2.3

    .. note::
        ``watch`` can be used with service.running to restart a service when
         another state changes ( example: a file.managed state that creates the
         service's config file ). More details regarding ``watch`` can be found
         in the :ref:`Requisites <requisites>` documentation.
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    # used to let execution modules know which service state is being run.
    __context__["service.state"] = "running"

    # Check for common error: using enabled option instead of enable
    if "enabled" in kwargs:
        return _enabled_used_error(ret)

    # Convert enable to boolean in case user passed a string value
    if isinstance(enable, str):
        enable = salt.utils.data.is_true(enable)

    if _offline():
        ret["result"] = True
        ret["comment"] = "Running in OFFLINE mode. Nothing to do"
        return ret

    # Check if the service is available
    try:
        if not _available(name, ret):
            if __opts__.get("test"):
                ret["result"] = None
                ret["comment"] = (
                    "Service {} not present; if created in this state run, "
                    "it would have been started".format(name)
                )
            return ret
    except CommandExecutionError as exc:
        ret["result"] = False
        ret["comment"] = exc.strerror
        return ret

    status_kwargs, warnings = _get_systemd_only(__salt__["service.status"], kwargs)
    if warnings:
        _add_warnings(ret, warnings)

    # lot of custom init script won't or mis implement the status
    # command, so it is just an indicator but can not be fully trusted
    before_toggle_status = __salt__["service.status"](name, sig, **status_kwargs)
    if "service.enabled" in __salt__:
        before_toggle_enable_status = __salt__["service.enabled"](name)
    else:
        before_toggle_enable_status = True

    unmask_ret = {"comment": ""}
    if kwargs.get("unmask", False):
        unmask_ret = unmasked(name, kwargs.get("unmask_runtime", False))

    # See if the service is already running
    if before_toggle_status:
        ret["comment"] = "\n".join(
            [
                _f
                for _f in [
                    "The service {} is already running".format(name),
                    unmask_ret["comment"],
                ]
                if _f
            ]
        )
        if enable is True and not before_toggle_enable_status:
            ret.update(_enable(name, None, **kwargs))
        elif enable is False and before_toggle_enable_status:
            ret.update(_disable(name, None, **kwargs))
        return ret

    # Run the tests
    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "\n".join(
            [
                _f
                for _f in [
                    "Service {} is set to start".format(name),
                    unmask_ret["comment"],
                ]
                if _f
            ]
        )
        return ret

    # Conditionally add systemd-specific args to call to service.start
    start_kwargs, warnings = _get_systemd_only(__salt__["service.start"], kwargs)
    if warnings:
        _add_warnings(ret, warnings)

    if salt.utils.platform.is_windows() and kwargs.get("timeout", False):
        start_kwargs.update({"timeout": kwargs.get("timeout")})

    # macOS and Windows services cant be started if they are disabled. So we need
    # to enable them prior to starting otherwise we will always fail.
    macos = salt.utils.platform.is_darwin()
    windows = salt.utils.platform.is_windows()
    if (macos or windows) and not before_toggle_enable_status:
        # fail if enable isn't set.
        if not enable:
            ret["comment"] = (
                "The service {} is disabled but enable is not True. Set "
                "enable to True to successfully start the service.".format(name)
            )
            ret["result"] = False
            return ret
        ret.update(_enable(name, None, **kwargs))
        # we've already enabled we don't need to enable again later.
        enable = None

    try:
        func_ret = __salt__["service.start"](name, **start_kwargs)
    except CommandExecutionError as exc:
        ret["result"] = False
        ret["comment"] = exc.strerror
        return ret

    if not func_ret:
        ret["result"] = False
        ret["comment"] = "Service {} failed to start".format(name)
        if enable is True:
            ret.update(_enable(name, False, result=False, **kwargs))
        elif enable is False:
            ret.update(_disable(name, False, result=False, **kwargs))
        return ret

    if init_delay:
        time.sleep(init_delay)

    # only force a change state if we have explicitly detected them
    after_toggle_status = __salt__["service.status"](name, sig, **status_kwargs)
    if "service.enabled" in __salt__:
        after_toggle_enable_status = __salt__["service.enabled"](name)
    else:
        after_toggle_enable_status = True
    if (
        (before_toggle_enable_status != after_toggle_enable_status)
        or (before_toggle_status != after_toggle_status)
    ) and not ret.get("changes", {}):
        ret["changes"][name] = after_toggle_status

    if after_toggle_status:
        ret["comment"] = "Started service {}".format(name)
    else:
        ret["comment"] = "Service {} failed to start".format(name)
        ret["result"] = False

    if enable is True:
        ret.update(
            _enable(name, after_toggle_status, result=after_toggle_status, **kwargs)
        )
    elif enable is False:
        ret.update(
            _disable(name, after_toggle_status, result=after_toggle_status, **kwargs)
        )

    if init_delay:
        ret["comment"] = "{}\nDelayed return for {} seconds".format(
            ret["comment"], init_delay
        )

    if kwargs.get("unmask", False):
        ret["comment"] = "\n".join([ret["comment"], unmask_ret["comment"]])

    return ret


def dead(name, enable=None, sig=None, init_delay=None, **kwargs):
    """
    Ensure that the named service is dead by stopping the service if it is running

    name
        The name of the init or rc script used to manage the service

    enable
        Set the service to be enabled at boot time, ``True`` sets the service
        to be enabled, ``False`` sets the named service to be disabled. The
        default is ``None``, which does not enable or disable anything.

    sig
        The string to search for when looking for the service process with ps

    init_delay
        Add a sleep command (in seconds) before the check to make sure service
        is killed.

        .. versionadded:: 2017.7.0

    no_block : False
        **For systemd minions only.** Stops the service using ``--no-block``.

        .. versionadded:: 2017.7.0

    timeout
        **For Windows minions only.**

        The time in seconds to wait for the service to stop before returning.
        Default is the default for :py:func:`win_service.stop
        <salt.modules.win_service.stop>`.

        .. versionadded:: 2017.7.9,2018.3.4

    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    # used to let execution modules know which service state is being run.
    __context__["service.state"] = "dead"

    # Check for common error: using enabled option instead of enable
    if "enabled" in kwargs:
        return _enabled_used_error(ret)

    # Convert enable to boolean in case user passed a string value
    if isinstance(enable, str):
        enable = salt.utils.data.is_true(enable)

    if _offline():
        ret["result"] = True
        ret["comment"] = "Running in OFFLINE mode. Nothing to do"
        return ret

    # Check if the service is available
    try:
        if not _available(name, ret):
            if __opts__.get("test"):
                ret["result"] = None
                ret["comment"] = (
                    "Service {} not present; if created in this state run, "
                    "it would have been stopped".format(name)
                )
            else:
                # A non-available service is OK here, don't let the state fail
                # because of it.
                ret["result"] = True
            return ret
    except CommandExecutionError as exc:
        ret["result"] = False
        ret["comment"] = exc.strerror
        return ret

    status_kwargs, warnings = _get_systemd_only(__salt__["service.status"], kwargs)
    if warnings:
        _add_warnings(ret, warnings)

    # lot of custom init script won't or mis implement the status
    # command, so it is just an indicator but can not be fully trusted
    before_toggle_status = __salt__["service.status"](name, sig, **status_kwargs)
    if "service.enabled" in __salt__:
        if salt.utils.platform.is_windows():
            # service.enabled in Windows returns True for services that are set
            # to Auto start, but services set to Manual can also be disabled
            before_toggle_enable_status = __salt__["service.info"](name)[
                "StartType"
            ] in ["Auto", "Manual"]
        else:
            before_toggle_enable_status = __salt__["service.enabled"](name)
    else:
        before_toggle_enable_status = True

    # See if the service is already dead
    if not before_toggle_status:
        ret["comment"] = "The service {} is already dead".format(name)
        if enable is True and not before_toggle_enable_status:
            ret.update(_enable(name, None, **kwargs))
        elif enable is False and before_toggle_enable_status:
            ret.update(_disable(name, None, **kwargs))
        return ret

    # Run the tests
    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "Service {} is set to be killed".format(name)
        return ret

    # Conditionally add systemd-specific args to call to service.start
    stop_kwargs, warnings = _get_systemd_only(__salt__["service.stop"], kwargs)
    if warnings:
        _add_warnings(ret, warnings)

    if salt.utils.platform.is_windows() and kwargs.get("timeout", False):
        stop_kwargs.update({"timeout": kwargs.get("timeout")})

    func_ret = __salt__["service.stop"](name, **stop_kwargs)
    if not func_ret:
        ret["result"] = False
        ret["comment"] = "Service {} failed to die".format(name)
        if enable is True:
            ret.update(_enable(name, True, result=False, **kwargs))
        elif enable is False:
            ret.update(_disable(name, True, result=False, **kwargs))
        return ret

    if init_delay:
        time.sleep(init_delay)

    # only force a change state if we have explicitly detected them
    after_toggle_status = __salt__["service.status"](name, **status_kwargs)
    if "service.enabled" in __salt__:
        after_toggle_enable_status = __salt__["service.enabled"](name)
    else:
        after_toggle_enable_status = True
    if (
        (before_toggle_enable_status != after_toggle_enable_status)
        or (before_toggle_status != after_toggle_status)
    ) and not ret.get("changes", {}):
        ret["changes"][name] = after_toggle_status

    # be sure to stop, in case we mis detected in the check
    if after_toggle_status:
        ret["result"] = False
        ret["comment"] = "Service {} failed to die".format(name)
    else:
        ret["comment"] = "Service {} was killed".format(name)

    if enable is True:
        ret.update(
            _enable(name, after_toggle_status, result=not after_toggle_status, **kwargs)
        )
    elif enable is False:
        ret.update(
            _disable(
                name, after_toggle_status, result=not after_toggle_status, **kwargs
            )
        )

    return ret


def enabled(name, **kwargs):
    """
    Ensure that the service is enabled on boot, only use this state if you
    don't want to manage the running process, remember that if you want to
    enable a running service to use the enable: True option for the running
    or dead function.

    name
        The name of the init or rc script used to manage the service
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    # used to let execution modules know which service state is being run.
    __context__["service.state"] = "enabled"

    ret.update(_enable(name, None, **kwargs))
    if __opts__.get("test") and ret.get(
        "comment"
    ) == "The named service {} is not available".format(name):
        ret["result"] = None
        ret["comment"] = (
            "Service {} not present; if created in this state run, "
            "it would have been enabled".format(name)
        )
    return ret


def disabled(name, **kwargs):
    """
    Ensure that the service is disabled on boot, only use this state if you
    don't want to manage the running process, remember that if you want to
    disable a service to use the enable: False option for the running or dead
    function.

    name
        The name of the init or rc script used to manage the service
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    # used to let execution modules know which service state is being run.
    __context__["service.state"] = "disabled"

    ret.update(_disable(name, None, **kwargs))
    return ret


def masked(name, runtime=False):
    """
    .. versionadded:: 2017.7.0

    .. note::
        This state is only available on minions which use systemd_.

    Ensures that the named service is masked (i.e. prevented from being
    started).

    name
        Name of the service to mask

    runtime : False
        By default, this state will manage an indefinite mask for the named
        service. Set this argument to ``True`` to runtime mask the service.

    .. note::
        It is possible for a service to have both indefinite and runtime masks
        set for it. Therefore, this state will manage a runtime or indefinite
        mask independently of each other. This means that if the service is
        already indefinitely masked, running this state with ``runtime`` set to
        ``True`` will _not_ remove the indefinite mask before setting a runtime
        mask. In these cases, if it is desirable to ensure that the service is
        runtime masked and not indefinitely masked, pair this state with a
        :py:func:`service.unmasked <salt.states.service.unmasked>` state, like
        so:

        .. code-block:: yaml

            mask_runtime_foo:
              service.masked:
                - name: foo
                - runtime: True

            unmask_indefinite_foo:
              service.unmasked:
                - name: foo
                - runtime: False

    .. _systemd: https://freedesktop.org/wiki/Software/systemd/

    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    if "service.masked" not in __salt__:
        ret["comment"] = "Service masking not available on this minion"
        ret["result"] = False
        return ret

    mask_type = "runtime masked" if runtime else "masked"
    expected_changes = {mask_type: {"old": False, "new": True}}

    try:
        if __salt__["service.masked"](name, runtime):
            ret["comment"] = "Service {} is already {}".format(
                name,
                mask_type,
            )
            return ret

        if __opts__["test"]:
            ret["result"] = None
            ret["changes"] = expected_changes
            ret["comment"] = "Service {} would be {}".format(name, mask_type)
            return ret

        __salt__["service.mask"](name, runtime)

        if __salt__["service.masked"](name, runtime):
            ret["changes"] = expected_changes
            ret["comment"] = "Service {} was {}".format(name, mask_type)
        else:
            ret["comment"] = "Failed to mask service {}".format(name)
        return ret

    except CommandExecutionError as exc:
        ret["result"] = False
        ret["comment"] = exc.strerror
        return ret


def unmasked(name, runtime=False):
    """
    .. versionadded:: 2017.7.0

    .. note::
        This state is only available on minions which use systemd_.

    Ensures that the named service is unmasked

    name
        Name of the service to unmask

    runtime : False
        By default, this state will manage an indefinite mask for the named
        service. Set this argument to ``True`` to ensure that the service is
        runtime masked.

    .. note::
        It is possible for a service to have both indefinite and runtime masks
        set for it. Therefore, this state will manage a runtime or indefinite
        mask independently of each other. This means that if the service is
        indefinitely masked, running this state with ``runtime`` set to
        ``True`` will _not_ remove the indefinite mask.

    .. _systemd: https://freedesktop.org/wiki/Software/systemd/

    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    if "service.masked" not in __salt__:
        ret["comment"] = "Service masking not available on this minion"
        ret["result"] = False
        return ret

    mask_type = "runtime masked" if runtime else "masked"
    action = "runtime unmasked" if runtime else "unmasked"
    expected_changes = {mask_type: {"old": True, "new": False}}

    try:
        if not __salt__["service.masked"](name, runtime):
            ret["comment"] = "Service {} was already {}".format(name, action)
            return ret

        if __opts__["test"]:
            ret["result"] = None
            ret["changes"] = expected_changes
            ret["comment"] = "Service {} would be {}".format(name, action)
            return ret

        __salt__["service.unmask"](name, runtime)

        if not __salt__["service.masked"](name, runtime):
            ret["changes"] = expected_changes
            ret["comment"] = "Service {} was {}".format(name, action)
        else:
            ret["comment"] = "Failed to unmask service {}".format(name)
        return ret

    except CommandExecutionError as exc:
        ret["result"] = False
        ret["comment"] = exc.strerror
        return ret


def mod_watch(
    name,
    sfun=None,
    sig=None,
    reload=False,
    full_restart=False,
    init_delay=None,
    force=False,
    **kwargs
):
    """
    The service watcher, called to invoke the watch command.
    When called, it will restart or reload the named service.

    .. note::
        This state exists to support special handling of the ``watch``
        :ref:`requisite <requisites>`. It should not be called directly.

        Parameters for this function should be set by the watching service
        (e.g. ``service.running``).

    name
        The name of the service to control.

    sfun
        The original function which triggered the mod_watch call
        (`service.running`, for example).

    sig
        The string to search for when looking for the service process with ps.

    reload
        When set, reload the service instead of restarting it
        (e.g. ``service nginx reload``).

    full_restart
        Perform a full stop/start of a service by passing ``--full-restart``.
        This option is ignored if ``reload`` is set and is supported by only a few
        :py:func:`service modules <salt.modules.service>`.

    force
        Use service.force_reload instead of reload (needs reload to be set to True).

    init_delay
        Add a sleep command (in seconds) before the service is restarted/reloaded.
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}
    past_participle = None

    status_kwargs, warnings = _get_systemd_only(__salt__["service.status"], kwargs)
    if warnings:
        _add_warnings(ret, warnings)

    if sfun == "dead":
        verb = "stop"
        past_participle = verb + "ped"
        if __salt__["service.status"](name, sig, **status_kwargs):
            func = __salt__["service.stop"]
        else:
            ret["result"] = True
            ret["comment"] = "Service is already {}".format(past_participle)
            return ret
    elif sfun == "running":
        if __salt__["service.status"](name, sig, **status_kwargs):
            if "service.reload" in __salt__ and reload:
                if "service.force_reload" in __salt__ and force:
                    func = __salt__["service.force_reload"]
                    verb = "forcefully reload"
                else:
                    func = __salt__["service.reload"]
                    verb = "reload"
            elif "service.full_restart" in __salt__ and full_restart:
                func = __salt__["service.full_restart"]
                verb = "fully restart"
            else:
                func = __salt__["service.restart"]
                verb = "restart"
        else:
            func = __salt__["service.start"]
            verb = "start"
        if not past_participle:
            past_participle = verb + "ed"
    else:
        ret["comment"] = "Unable to trigger watch for service.{}".format(sfun)
        ret["result"] = False
        return ret

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "Service is set to be {}".format(past_participle)
        return ret

    if verb == "start" and "service.stop" in __salt__:
        # stop service before start
        __salt__["service.stop"](name)

    func_kwargs, warnings = _get_systemd_only(func, kwargs)
    if warnings:
        _add_warnings(ret, warnings)

    try:
        result = func(name, **func_kwargs)
    except CommandExecutionError as exc:
        ret["result"] = False
        ret["comment"] = exc.strerror
        return ret

    if init_delay:
        time.sleep(init_delay)

    ret["changes"] = {name: result}
    ret["result"] = result
    ret["comment"] = (
        "Service {}".format(past_participle)
        if result
        else "Failed to {} the service".format(verb)
    )
    return ret


def mod_beacon(name, **kwargs):
    """
    Create a beacon to monitor a service based on a beacon state argument.

    .. note::
        This state exists to support special handling of the ``beacon``
        state argument for supported state functions. It should not be called directly.
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    sfun = kwargs.pop("sfun", None)
    supported_funcs = ["running", "dead"]

    if sfun in supported_funcs:
        if kwargs.get("beacon"):
            beacon_module = "service"

            data = {}
            _beacon_data = kwargs.get("beacon_data", {})

            data["onchangeonly"] = _beacon_data.get("onchangeonly", True)
            data["delay"] = _beacon_data.get("delay", 0)
            data["emitatstartup"] = _beacon_data.get("emitatstartup", False)
            data["uncleanshutdown"] = _beacon_data.get("emitatstartup", None)

            beacon_name = "beacon_{}_{}".format(beacon_module, name)

            beacon_kwargs = {
                "name": beacon_name,
                "services": {name: data},
                "interval": _beacon_data.get("interval", 60),
                "beacon_module": beacon_module,
            }

            ret = __states__["beacon.present"](**beacon_kwargs)
            return ret
        else:
            return {
                "name": name,
                "changes": {},
                "comment": "Not adding beacon.",
                "result": True,
            }
    else:
        return {
            "name": name,
            "changes": {},
            "comment": "service.{} does not work with the beacon state function".format(
                sfun
            ),
            "result": False,
        }



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    Test cases for salt.modules.incron
"""


import pytest

import salt.modules.incron as incron
from tests.support.mock import MagicMock, call, patch


@pytest.fixture
def configure_loader_modules():
    return {incron: {}}


# 'write_incron_file' function tests: 1


def test_write_incron_file():
    """
    Test if it writes the contents of a file to a user's crontab
    """
    mock = MagicMock(return_value=0)
    with patch.dict(incron.__salt__, {"cmd.retcode": mock}), patch(
        "salt.modules.incron._get_incron_cmdstr",
        MagicMock(return_value="incrontab"),
    ):
        assert incron.write_incron_file("cybage", "/home/cybage/new_cron")


# 'write_cron_file_verbose' function tests: 1


def test_write_cron_file_verbose():
    """
    Test if it writes the contents of a file to a user's crontab and
    return error message on error
    """
    mock = MagicMock(return_value=True)
    with patch.dict(incron.__salt__, {"cmd.run_all": mock}), patch(
        "salt.modules.incron._get_incron_cmdstr",
        MagicMock(return_value="incrontab"),
    ):
        assert incron.write_incron_file_verbose("cybage", "/home/cybage/new_cron")


# 'raw_system_incron' function tests: 1


def test_raw_system_incron():
    """
    Test if it return the contents of the system wide incrontab
    """
    with patch("salt.modules.incron._read_file", MagicMock(return_value="salt")):
        assert incron.raw_system_incron() == "salt"


# 'raw_incron' function tests: 1


def test_raw_incron():
    """
    Test if it return the contents of the user's incrontab
    """
    mock = MagicMock(return_value="incrontab")
    expected_calls = [
        call("incrontab -l cybage", python_shell=False, rstrip=False, runas="cybage")
    ]

    with patch.dict(incron.__grains__, {"os_family": mock}):
        cmd_run_mock = MagicMock(return_value="salt")
        with patch.dict(incron.__salt__, {"cmd.run_stdout": cmd_run_mock}):
            assert incron.raw_incron("cybage") == "salt"
            cmd_run_mock.assert_has_calls(expected_calls)

            cmd = cmd_run_mock.call_args[0][0]
            assert "incrontab -l cybage" == cmd
            assert "-u" not in cmd


# 'list_tab' function tests: 1


def test_list_tab():
    """
    Test if it return the contents of the specified user's incrontab
    """
    mock = MagicMock(return_value="incrontab")
    with patch.dict(incron.__grains__, {"os_family": mock}):
        mock = MagicMock(return_value="salt")
        with patch.dict(incron.__salt__, {"cmd.run_stdout": mock}):
            assert incron.list_tab("cybage") == {"pre": ["salt"], "crons": []}


# 'set_job' function tests: 1


def test_set_job():
    """
    Test if it sets a cron job up for a specified user.
    """
    assert (
        incron.set_job("cybage", "/home/cybage", "TO_MODIFY", 'echo "$$ $@ $# $% $&"')
        == "Invalid mask type: TO_MODIFY"
    )

    val = {
        "pre": [],
        "crons": [{"path": "/home/cybage", "mask": "IN_MODIFY", "cmd": 'echo "SALT"'}],
    }
    with patch.object(incron, "list_tab", MagicMock(return_value=val)):
        assert (
            incron.set_job("cybage", "/home/cybage", "IN_MODIFY", 'echo "SALT"')
            == "present"
        )

    with patch.object(
        incron, "list_tab", MagicMock(return_value={"pre": ["salt"], "crons": []})
    ):
        mock = MagicMock(return_value="incrontab")
        with patch.dict(incron.__grains__, {"os_family": mock}):
            with patch.object(
                incron,
                "_write_incron_lines",
                MagicMock(return_value={"retcode": True, "stderr": "error"}),
            ):
                assert (
                    incron.set_job("cybage", "/home/cybage", "IN_MODIFY", 'echo "SALT"')
                    == "error"
                )

    with patch.object(
        incron, "list_tab", MagicMock(return_value={"pre": ["salt"], "crons": []})
    ):
        mock = MagicMock(return_value="incrontab")
        with patch.dict(incron.__grains__, {"os_family": mock}):
            with patch.object(
                incron,
                "_write_incron_lines",
                MagicMock(return_value={"retcode": False, "stderr": "error"}),
            ):
                assert (
                    incron.set_job("cybage", "/home/cybage", "IN_MODIFY", 'echo "SALT"')
                    == "new"
                )

    val = {
        "pre": [],
        "crons": [
            {
                "path": "/home/cybage",
                "mask": "IN_MODIFY,IN_DELETE",
                "cmd": 'echo "SALT"',
            }
        ],
    }
    with patch.object(incron, "list_tab", MagicMock(return_value=val)):
        mock = MagicMock(return_value="incrontab")
        with patch.dict(incron.__grains__, {"os_family": mock}):
            with patch.object(
                incron,
                "_write_incron_lines",
                MagicMock(return_value={"retcode": False, "stderr": "error"}),
            ):
                assert (
                    incron.set_job("cybage", "/home/cybage", "IN_DELETE", 'echo "SALT"')
                    == "updated"
                )


# 'rm_job' function tests: 1


def test_rm_job():
    """
    Test if it remove a cron job for a specified user. If any of the
    day/time params are specified, the job will only be removed if
    the specified params match.
    """
    assert (
        incron.rm_job("cybage", "/home/cybage", "TO_MODIFY", 'echo "$$ $@ $# $% $&"')
        == "Invalid mask type: TO_MODIFY"
    )

    with patch.object(
        incron, "list_tab", MagicMock(return_value={"pre": ["salt"], "crons": []})
    ):
        mock = MagicMock(return_value="incrontab")
        with patch.dict(incron.__grains__, {"os_family": mock}):
            with patch.object(
                incron,
                "_write_incron_lines",
                MagicMock(return_value={"retcode": True, "stderr": "error"}),
            ):
                assert (
                    incron.rm_job("cybage", "/home/cybage", "IN_MODIFY", 'echo "SALT"')
                    == "error"
                )

    with patch.object(
        incron, "list_tab", MagicMock(return_value={"pre": ["salt"], "crons": []})
    ):
        mock = MagicMock(return_value="incrontab")
        with patch.dict(incron.__grains__, {"os_family": mock}):
            with patch.object(
                incron,
                "_write_incron_lines",
                MagicMock(return_value={"retcode": False, "stderr": "error"}),
            ):
                assert (
                    incron.rm_job("cybage", "/home/cybage", "IN_MODIFY", 'echo "SALT"')
                    == "absent"
                )



"""
Module for managing timezone on Windows systems.
"""
import logging
from datetime import datetime

from salt.exceptions import CommandExecutionError

try:
    import pytz

    HAS_PYTZ = True
except ImportError:
    HAS_PYTZ = False

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "timezone"


class TzMapper:
    def __init__(self, unix_to_win):
        self.win_to_unix = {k.lower(): v for k, v in unix_to_win.items()}
        self.unix_to_win = {v.lower(): k for k, v in unix_to_win.items()}

    def add(self, k, v):
        self.unix_to_win[k.lower()] = v
        self.win_to_unix[v.lower()] = k

    def remove(self, k):
        self.win_to_unix.pop(self.unix_to_win.pop(k.lower()).lower())

    def get_win(self, key, default=None):
        return self.unix_to_win.get(key.lower(), default)

    def get_unix(self, key, default=None):
        return self.win_to_unix.get(key.lower(), default)

    def list_win(self):
        return sorted(self.unix_to_win.values())

    def list_unix(self):
        return sorted(self.win_to_unix.values())


mapper = TzMapper(
    {
        "AUS Central Standard Time": "Australia/Darwin",
        "AUS Eastern Standard Time": "Australia/Sydney",
        "Afghanistan Standard Time": "Asia/Kabul",
        "Alaskan Standard Time": "America/Anchorage",
        "Aleutian Standard Time": "America/Adak",
        "Altai Standard Time": "Asia/Barnaul",
        "Arab Standard Time": "Asia/Riyadh",
        "Arabian Standard Time": "Asia/Dubai",
        "Arabic Standard Time": "Asia/Baghdad",
        "Argentina Standard Time": "America/Buenos_Aires",
        "Astrakhan Standard Time": "Europe/Astrakhan",
        "Atlantic Standard Time": "America/Halifax",
        "Aus Central W. Standard Time": "Australia/Eucla",
        "Azerbaijan Standard Time": "Asia/Baku",
        "Azores Standard Time": "Atlantic/Azores",
        "Bahia Standard Time": "America/Bahia",
        "Bangladesh Standard Time": "Asia/Dhaka",
        "Belarus Standard Time": "Europe/Minsk",
        "Bougainville Standard Time": "Pacific/Bougainville",
        "Canada Central Standard Time": "America/Regina",
        "Cape Verde Standard Time": "Atlantic/Cape_Verde",
        "Caucasus Standard Time": "Asia/Yerevan",
        "Cen. Australia Standard Time": "Australia/Adelaide",
        "Central America Standard Time": "America/Guatemala",
        "Central Asia Standard Time": "Asia/Almaty",
        "Central Brazilian Standard Time": "America/Cuiaba",
        "Central Europe Standard Time": "Europe/Budapest",
        "Central European Standard Time": "Europe/Warsaw",
        "Central Pacific Standard Time": "Pacific/Guadalcanal",
        "Central Standard Time": "America/Chicago",
        "Central Standard Time (Mexico)": "America/Mexico_City",
        "Chatham Islands Standard Time": "Pacific/Chatham",
        "China Standard Time": "Asia/Shanghai",
        "Cuba Standard Time": "America/Havana",
        "Dateline Standard Time": "Etc/GMT+12",
        "E. Africa Standard Time": "Africa/Nairobi",
        "E. Australia Standard Time": "Australia/Brisbane",
        "E. Europe Standard Time": "Europe/Chisinau",
        "E. South America Standard Time": "America/Sao_Paulo",
        "Easter Island Standard Time": "Pacific/Easter",
        "Eastern Standard Time": "America/New_York",
        "Eastern Standard Time (Mexico)": "America/Cancun",
        "Egypt Standard Time": "Africa/Cairo",
        "Ekaterinburg Standard Time": "Asia/Yekaterinburg",
        "FLE Standard Time": "Europe/Kiev",
        "Fiji Standard Time": "Pacific/Fiji",
        "GMT Standard Time": "Europe/London",
        "GTB Standard Time": "Europe/Bucharest",
        "Georgian Standard Time": "Asia/Tbilisi",
        "Greenland Standard Time": "America/Godthab",
        "Greenwich Standard Time": "Atlantic/Reykjavik",
        "Haiti Standard Time": "America/Port-au-Prince",
        "Hawaiian Standard Time": "Pacific/Honolulu",
        "India Standard Time": "Asia/Calcutta",
        "Iran Standard Time": "Asia/Tehran",
        "Israel Standard Time": "Asia/Jerusalem",
        "Jordan Standard Time": "Asia/Amman",
        "Kaliningrad Standard Time": "Europe/Kaliningrad",
        "Korea Standard Time": "Asia/Seoul",
        "Libya Standard Time": "Africa/Tripoli",
        "Line Islands Standard Time": "Pacific/Kiritimati",
        "Lord Howe Standard Time": "Australia/Lord_Howe",
        "Magadan Standard Time": "Asia/Magadan",
        "Magallanes Standard Time": "America/Punta_Arenas",
        "Marquesas Standard Time": "Pacific/Marquesas",
        "Mauritius Standard Time": "Indian/Mauritius",
        "Middle East Standard Time": "Asia/Beirut",
        "Montevideo Standard Time": "America/Montevideo",
        "Morocco Standard Time": "Africa/Casablanca",
        "Mountain Standard Time": "America/Denver",
        "Mountain Standard Time (Mexico)": "America/Chihuahua",
        "Myanmar Standard Time": "Asia/Rangoon",
        "N. Central Asia Standard Time": "Asia/Novosibirsk",
        "Namibia Standard Time": "Africa/Windhoek",
        "Nepal Standard Time": "Asia/Katmandu",
        "New Zealand Standard Time": "Pacific/Auckland",
        "Newfoundland Standard Time": "America/St_Johns",
        "Norfolk Standard Time": "Pacific/Norfolk",
        "North Asia East Standard Time": "Asia/Irkutsk",
        "North Asia Standard Time": "Asia/Krasnoyarsk",
        "North Korea Standard Time": "Asia/Pyongyang",
        "Omsk Standard Time": "Asia/Omsk",
        "Pacific SA Standard Time": "America/Santiago",
        "Pacific Standard Time": "America/Los_Angeles",
        "Pacific Standard Time (Mexico)": "America/Tijuana",
        "Pakistan Standard Time": "Asia/Karachi",
        "Paraguay Standard Time": "America/Asuncion",
        "Romance Standard Time": "Europe/Paris",
        "Russia Time Zone 10": "Asia/Srednekolymsk",
        "Russia Time Zone 11": "Asia/Kamchatka",
        "Russia Time Zone 3": "Europe/Samara",
        "Russian Standard Time": "Europe/Moscow",
        "SA Eastern Standard Time": "America/Cayenne",
        "SA Pacific Standard Time": "America/Bogota",
        "SA Western Standard Time": "America/La_Paz",
        "SE Asia Standard Time": "Asia/Bangkok",
        "Saint Pierre Standard Time": "America/Miquelon",
        "Sakhalin Standard Time": "Asia/Sakhalin",
        "Samoa Standard Time": "Pacific/Apia",
        "Saratov Standard Time": "Europe/Saratov",
        "Singapore Standard Time": "Asia/Singapore",
        "South Africa Standard Time": "Africa/Johannesburg",
        "Sri Lanka Standard Time": "Asia/Colombo",
        "Syria Standard Time": "Asia/Damascus",
        "Taipei Standard Time": "Asia/Taipei",
        "Tasmania Standard Time": "Australia/Hobart",
        "Tocantins Standard Time": "America/Araguaina",
        "Tokyo Standard Time": "Asia/Tokyo",
        "Tomsk Standard Time": "Asia/Tomsk",
        "Tonga Standard Time": "Pacific/Tongatapu",
        "Transbaikal Standard Time": "Asia/Chita",
        "Turkey Standard Time": "Europe/Istanbul",
        "Turks And Caicos Standard Time": "America/Grand_Turk",
        "US Eastern Standard Time": "America/Indianapolis",
        "US Mountain Standard Time": "America/Phoenix",
        "UTC": "Etc/GMT",
        "UTC+12": "Etc/GMT-12",
        "UTC+13": "Etc/GMT-13",
        "UTC-02": "Etc/GMT+2",
        "UTC-08": "Etc/GMT+8",
        "UTC-09": "Etc/GMT+9",
        "UTC-11": "Etc/GMT+11",
        "Ulaanbaatar Standard Time": "Asia/Ulaanbaatar",
        "Venezuela Standard Time": "America/Caracas",
        "Vladivostok Standard Time": "Asia/Vladivostok",
        "W. Australia Standard Time": "Australia/Perth",
        "W. Central Africa Standard Time": "Africa/Lagos",
        "W. Europe Standard Time": "Europe/Berlin",
        "W. Mongolia Standard Time": "Asia/Hovd",
        "West Asia Standard Time": "Asia/Tashkent",
        "West Bank Standard Time": "Asia/Hebron",
        "West Pacific Standard Time": "Pacific/Port_Moresby",
        "Yakutsk Standard Time": "Asia/Yakutsk",
    }
)


def __virtual__():
    """
    Only load on windows
    """
    if not __utils__["platform.is_windows"]():
        return False, "Module win_timezone: Not on Windows client"
    if not HAS_PYTZ:
        return False, "Module win_timezone: pytz not found"
    if not __utils__["path.which"]("tzutil"):
        return False, "Module win_timezone: tzutil not found"
    return __virtualname__


def get_zone():
    """
    Get current timezone (i.e. America/Denver)

    Returns:
        str: Timezone in unix format

    Raises:
        CommandExecutionError: If timezone could not be gathered

    CLI Example:

    .. code-block:: bash

        salt '*' timezone.get_zone
    """
    cmd = ["tzutil", "/g"]
    res = __salt__["cmd.run_all"](cmd, python_shell=False)
    if res["retcode"] or not res["stdout"]:
        raise CommandExecutionError(
            "tzutil encountered an error getting timezone", info=res
        )
    tz = res["stdout"].lower()
    if tz.endswith("_dstoff"):
        tz = tz[:-7]

    return mapper.get_unix(tz, "Unknown")


def get_offset():
    """
    Get current numeric timezone offset from UTC (i.e. -0700)

    Returns:
        str: Offset from UTC

    CLI Example:

    .. code-block:: bash

        salt '*' timezone.get_offset
    """
    # http://craigglennie.com/programming/python/2013/07/21/working-with-timezones-using-Python-and-pytz-localize-vs-normalize/
    tz_object = pytz.timezone(get_zone())
    utc_time = pytz.utc.localize(datetime.utcnow())
    loc_time = utc_time.astimezone(tz_object)
    norm_time = tz_object.normalize(loc_time)
    return norm_time.strftime("%z")


def get_zonecode():
    """
    Get current timezone (i.e. PST, MDT, etc)

    Returns:
        str: An abbreviated timezone code

    CLI Example:

    .. code-block:: bash

        salt '*' timezone.get_zonecode
    """
    tz_object = pytz.timezone(get_zone())
    loc_time = tz_object.localize(datetime.utcnow())
    return loc_time.tzname()


def set_zone(timezone):
    """
    Sets the timezone using the tzutil.

    Args:
        timezone (str): A valid timezone

    Returns:
        bool: ``True`` if successful, otherwise ``False``

    Raises:
        CommandExecutionError: If invalid timezone is passed

    CLI Example:

    .. code-block:: bash

        salt '*' timezone.set_zone 'America/Denver'
    """
    # if it's one of the key's just use it
    if timezone.lower() in mapper.win_to_unix:
        win_zone = timezone

    elif timezone.lower() in mapper.unix_to_win:
        # if it's one of the values, use the key
        win_zone = mapper.get_win(timezone)

    else:
        # Raise error because it's neither key nor value
        raise CommandExecutionError("Invalid timezone passed: {}".format(timezone))

    # Set the value
    cmd = ["tzutil", "/s", win_zone]
    res = __salt__["cmd.run_all"](cmd, python_shell=False)
    if res["retcode"]:
        raise CommandExecutionError(
            "tzutil encountered an error setting timezone: {}".format(timezone),
            info=res,
        )
    return zone_compare(timezone)


def zone_compare(timezone):
    """
    Compares the given timezone with the machine timezone. Mostly useful for
    running state checks.

    Args:
        timezone (str):
            The timezone to compare. This can be in Windows or Unix format. Can
            be any of the values returned by the ``timezone.list`` function

    Returns:
        bool: ``True`` if they match, otherwise ``False``

    Example:

    .. code-block:: bash

        salt '*' timezone.zone_compare 'America/Denver'
    """
    # if it's one of the key's just use it
    if timezone.lower() in mapper.win_to_unix:
        check_zone = timezone

    elif timezone.lower() in mapper.unix_to_win:
        # if it's one of the values, use the key
        check_zone = mapper.get_win(timezone)

    else:
        # Raise error because it's neither key nor value
        raise CommandExecutionError("Invalid timezone passed: {}".format(timezone))

    return get_zone() == mapper.get_unix(check_zone, "Unknown")


def list(unix_style=True):
    """
    Return a list of Timezones that this module supports. These can be in either
    Unix or Windows format.

    .. versionadded:: 2018.3.3

    Args:
        unix_style (bool):
            ``True`` returns Unix-style timezones. ``False`` returns
            Windows-style timezones. Default is ``True``

    Returns:
        list: A list of supported timezones

    CLI Example:

    .. code-block:: bash

        # Unix-style timezones
        salt '*' timezone.list

        # Windows-style timezones
        salt '*' timezone.list unix_style=False
    """
    if unix_style:
        return mapper.list_unix()
    else:
        return mapper.list_win()


def get_hwclock():
    """
    Get current hardware clock setting (UTC or localtime)

    .. note::
        The hardware clock is always local time on Windows so this will always
        return "localtime"

    CLI Example:

    .. code-block:: bash

        salt '*' timezone.get_hwclock
    """
    # The hardware clock is always localtime on Windows
    return "localtime"


def set_hwclock(clock):
    """
    Sets the hardware clock to be either UTC or localtime

    .. note::
        The hardware clock is always local time on Windows so this will always
        return ``False``

    CLI Example:

    .. code-block:: bash

        salt '*' timezone.set_hwclock UTC
    """
    # The hardware clock is always localtime on Windows
    return False



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.lvm as lvm
from salt.exceptions import ArgumentValueError
from tests.support.mock import MagicMock, patch


@pytest.fixture
def lv01():
    return {
        "/dev/testvg01/testlv01": {
            "Logical Volume Name": "/dev/testvg01/testlv01",
            "Volume Group Name": "testvg01",
            "Logical Volume Access": "3",
            "Logical Volume Status": "1",
            "Internal Logical Volume Number": "-1",
            "Open Logical Volumes": "0",
            "Logical Volume Size": "4194304",
            "Current Logical Extents Associated": "512",
            "Allocated Logical Extents": "-1",
            "Allocation Policy": "0",
            "Read Ahead Sectors": "-1",
            "Major Device Number": "253",
            "Minor Device Number": "9",
        }
    }


@pytest.fixture
def lv02():
    return {
        "/dev/testvg01/testlv02": {
            "Logical Volume Name": "/dev/testvg01/testlv02",
            "Volume Group Name": "testvg01",
            "Logical Volume Access": "3",
            "Logical Volume Status": "1",
            "Internal Logical Volume Number": "-1",
            "Open Logical Volumes": "0",
            "Logical Volume Size": "4194304",
            "Current Logical Extents Associated": "512",
            "Allocated Logical Extents": "-1",
            "Allocation Policy": "0",
            "Read Ahead Sectors": "-1",
            "Major Device Number": "253",
            "Minor Device Number": "9",
        }
    }


@pytest.fixture
def configure_loader_modules():
    return {lvm: {}}


def test_pv_present():
    """
    Test to set a physical device to be used as an LVM physical volume
    """
    name = "/dev/sda5"

    comt = "Physical Volume {} already present".format(name)

    ret = {"name": name, "changes": {}, "result": True, "comment": comt}

    mock = MagicMock(side_effect=[True, False])
    with patch.dict(lvm.__salt__, {"lvm.pvdisplay": mock}):
        assert lvm.pv_present(name) == ret

        comt = "Physical Volume {} is set to be created".format(name)
        ret.update({"comment": comt, "result": None})
        with patch.dict(lvm.__opts__, {"test": True}):
            assert lvm.pv_present(name) == ret


def test_pv_absent():
    """
    Test to ensure that a Physical Device is not being used by lvm
    """
    name = "/dev/sda5"

    comt = "Physical Volume {} does not exist".format(name)

    ret = {"name": name, "changes": {}, "result": True, "comment": comt}

    mock = MagicMock(side_effect=[False, True])
    with patch.dict(lvm.__salt__, {"lvm.pvdisplay": mock}):
        assert lvm.pv_absent(name) == ret

        comt = "Physical Volume {} is set to be removed".format(name)
        ret.update({"comment": comt, "result": None})
        with patch.dict(lvm.__opts__, {"test": True}):
            assert lvm.pv_absent(name) == ret


def test_vg_present():
    """
    Test to create an LVM volume group
    """
    name = "testvg00"

    comt = "Failed to create Volume Group {}".format(name)

    ret = {"name": name, "changes": {}, "result": False, "comment": comt}

    mock = MagicMock(return_value=False)
    with patch.dict(lvm.__salt__, {"lvm.vgdisplay": mock, "lvm.vgcreate": mock}):
        with patch.dict(lvm.__opts__, {"test": False}):
            assert lvm.vg_present(name) == ret

        comt = "Volume Group {} is set to be created".format(name)
        ret.update({"comment": comt, "result": None})
        with patch.dict(lvm.__opts__, {"test": True}):
            assert lvm.vg_present(name) == ret


def test_vg_absent():
    """
    Test to remove an LVM volume group
    """
    name = "testvg00"

    comt = "Volume Group {} already absent".format(name)

    ret = {"name": name, "changes": {}, "result": True, "comment": comt}

    mock = MagicMock(side_effect=[False, True])
    with patch.dict(lvm.__salt__, {"lvm.vgdisplay": mock}):
        assert lvm.vg_absent(name) == ret

        comt = "Volume Group {} is set to be removed".format(name)
        ret.update({"comment": comt, "result": None})
        with patch.dict(lvm.__opts__, {"test": True}):
            assert lvm.vg_absent(name) == ret


def test_lv_present(lv01, lv02):
    """
    Test to create a new logical volume
    """
    name = "testlv01"
    vgname = "testvg01"
    comt = "Logical Volume {} already present".format(name)
    ret = {"name": name, "changes": {}, "result": True, "comment": comt}

    mock = MagicMock(return_value=lv01)
    with patch.dict(lvm.__salt__, {"lvm.lvdisplay": mock}):
        assert lvm.lv_present(name, vgname=vgname) == ret

    mock = MagicMock(return_value=lv02)
    with patch.dict(lvm.__salt__, {"lvm.lvdisplay": mock}):
        comt = "Logical Volume {} is set to be created".format(name)
        ret.update({"comment": comt, "result": None})
        with patch.dict(lvm.__opts__, {"test": True}):
            assert lvm.lv_present(name, vgname=vgname) == ret


def test_lv_present_with_valid_suffixes(lv01, lv02):
    """
    Test to create a new logical volume specifying valid suffixes
    """
    name = "testlv01"
    vgname = "testvg01"
    sizes_list = [
        "2048",
        "2048m",
        "2048M",
        "2048MB",
        "2048mb",
        "2g",
        "2G",
        "2GB",
        "2gb",
        "4194304s",
        "4194304S",
    ]
    comt = "Logical Volume {} already present".format(name)
    ret = {"name": name, "changes": {}, "result": True, "comment": comt}

    mock = MagicMock(return_value=lv01)
    with patch.dict(lvm.__salt__, {"lvm.lvdisplay": mock}):
        for size in sizes_list:
            assert lvm.lv_present(name, vgname=vgname, size=size) == ret

    sizes_list = [
        "1G",
        "1g",
        "2M",
        "2m",
        "3T",
        "3t",
        "4P",
        "4p",
        "5s",
        "5S",
        "1GB",
        "1gb",
        "2MB",
        "2mb",
        "3TB",
        "3tb",
        "4PB",
        "4pb",
    ]
    mock = MagicMock(return_value=lv02)
    with patch.dict(lvm.__salt__, {"lvm.lvdisplay": mock}):
        comt = "Logical Volume {} is set to be created".format(name)
        ret.update({"comment": comt, "result": None})
        with patch.dict(lvm.__opts__, {"test": True}):
            for size in sizes_list:
                assert lvm.lv_present(name, vgname=vgname, size=size) == ret


def test_lv_present_with_invalid_suffixes(lv02):
    """
    Test to create a new logical volume specifying valid suffixes
    """
    name = "testlv01"
    vgname = "testvg01"
    sizes_list = ["1B", "1b", "2K", "2k", "2KB", "2kb", "3BB", "3Bb", "4JKL", "YJK"]
    comt = "Logical Volume {} already present".format(name)
    ret = {"name": name, "changes": {}, "result": True, "comment": comt}

    mock = MagicMock(return_value=lv02)
    with patch.dict(lvm.__salt__, {"lvm.lvdisplay": mock}):
        comt = "Logical Volume {} is set to be created".format(name)
        ret.update({"comment": comt, "result": None})
        with patch.dict(lvm.__opts__, {"test": True}):
            for size in sizes_list:
                pytest.raises(
                    ArgumentValueError,
                    lvm.lv_present,
                    name,
                    vgname=vgname,
                    size=size,
                )


def test_lv_present_with_percentage_extents(lv01, lv02):
    """
    Test to create a new logical volume specifying extents as a percentage
    """
    name = "testlv01"
    vgname = "testvg01"
    extents = "42%FREE"
    comt = "Logical Volume {} already present, {} won't be resized.".format(
        name, extents
    )
    ret = {"name": name, "changes": {}, "result": True, "comment": comt}

    mock = MagicMock(return_value=lv01)
    with patch.dict(lvm.__salt__, {"lvm.lvdisplay": mock}):
        assert lvm.lv_present(name, vgname=vgname, extents=extents) == ret

    extents = "42%VG"
    mock = MagicMock(return_value=lv02)
    with patch.dict(lvm.__salt__, {"lvm.lvdisplay": mock}):
        comt = "Logical Volume {} is set to be created".format(name)
        ret.update({"comment": comt, "result": None})
        with patch.dict(lvm.__opts__, {"test": True}):
            assert lvm.lv_present(name, vgname=vgname, extents=extents) == ret


def test_lv_present_with_force(lv01, lv02):
    """
    Test to create a new logical volume with force=True
    """
    name = "testlv01"
    vgname = "testvg01"
    comt = "Logical Volume {} already present".format(name)
    ret = {"name": name, "changes": {}, "result": True, "comment": comt}

    mock = MagicMock(return_value=lv01)
    with patch.dict(lvm.__salt__, {"lvm.lvdisplay": mock}):
        assert lvm.lv_present(name, vgname=vgname, force=True) == ret

    mock = MagicMock(return_value=lv02)
    with patch.dict(lvm.__salt__, {"lvm.lvdisplay": mock}):
        comt = "Logical Volume {} is set to be created".format(name)
        ret.update({"comment": comt, "result": None})
        with patch.dict(lvm.__opts__, {"test": True}):
            assert lvm.lv_present(name, vgname=vgname, force=True) == ret


def test_lv_present_with_same_size(lv01):
    """
    Test to specify the same volume size as parameter
    """
    name = "testlv01"
    vgname = "testvg01"
    comt = "Logical Volume {} already present".format(name)
    ret = {"name": name, "changes": {}, "result": True, "comment": comt}

    mock = MagicMock(return_value=lv01)
    with patch.dict(lvm.__salt__, {"lvm.lvdisplay": mock}):
        assert lvm.lv_present(name, vgname=vgname, size="2G") == ret


def test_lv_present_with_increase(lv01):
    """
    Test to increase a logical volume
    """
    name = "testlv01"
    vgname = "testvg01"
    comt = "Logical Volume {} is set to be resized".format(name)
    ret = {"name": name, "changes": {}, "result": None, "comment": comt}

    mock = MagicMock(return_value=lv01)
    with patch.dict(lvm.__salt__, {"lvm.lvdisplay": mock}):
        with patch.dict(lvm.__opts__, {"test": True}):
            assert lvm.lv_present(name, vgname=vgname, size="10G") == ret


def test_lv_present_with_reduce_without_force(lv01):
    """
    Test to reduce a logical volume
    """
    name = "testlv01"
    vgname = "testvg01"
    comt = "To reduce a Logical Volume option 'force' must be True."
    ret = {"name": name, "changes": {}, "result": False, "comment": comt}

    mock = MagicMock(return_value=lv01)
    with patch.dict(lvm.__salt__, {"lvm.lvdisplay": mock}):
        assert lvm.lv_present(name, vgname=vgname, size="1G") == ret


def test_lv_present_with_reduce_with_force(lv01):
    """
    Test to reduce a logical volume
    """
    name = "testlv01"
    vgname = "testvg01"
    comt = "Logical Volume {} is set to be resized".format(name)
    ret = {"name": name, "changes": {}, "result": None, "comment": comt}

    mock = MagicMock(return_value=lv01)
    with patch.dict(lvm.__salt__, {"lvm.lvdisplay": mock}):
        with patch.dict(lvm.__opts__, {"test": True}):
            assert lvm.lv_present(name, vgname=vgname, size="1G", force=True) == ret


def test_lv_absent():
    """
    Test to remove a given existing logical volume from a named existing volume group
    """
    name = "testlv00"

    comt = "Logical Volume {} already absent".format(name)

    ret = {"name": name, "changes": {}, "result": True, "comment": comt}

    mock = MagicMock(side_effect=[False, True])
    with patch.dict(lvm.__salt__, {"lvm.lvdisplay": mock}):
        assert lvm.lv_absent(name) == ret

        comt = "Logical Volume {} is set to be removed".format(name)
        ret.update({"comment": comt, "result": None})
        with patch.dict(lvm.__opts__, {"test": True}):
            assert lvm.lv_absent(name) == ret



import json
import logging
import time

import pytest

import salt.utils.path
from tests.support.runtests import RUNTIME_VARS

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_if_binaries_missing("dockerd", "vault", "getent"),
]

VAULT_BINARY = salt.utils.path.which("vault")

log = logging.getLogger(__name__)


@pytest.fixture(scope="module")
def minion_config_overrides(vault_port):
    return {
        "vault": {
            "url": "http://127.0.0.1:{}".format(vault_port),
            "auth": {
                "method": "token",
                "token": "testsecret",
                "uses": 0,
                "policies": [
                    "testpolicy",
                ],
            },
        }
    }


def vault_container_version_id(value):
    return "vault=={}".format(value)


@pytest.fixture(
    scope="module",
    params=["0.9.6", "1.3.1", "latest"],
    ids=vault_container_version_id,
)
def vault_container_version(request, salt_factories, vault_port, shell):
    vault_version = request.param
    config = {
        "backend": {"file": {"path": "/vault/file"}},
        "default_lease_ttl": "168h",
        "max_lease_ttl": "720h",
        "disable_mlock": False,
    }

    factory = salt_factories.get_container(
        "vault",
        "ghcr.io/saltstack/salt-ci-containers/vault:{}".format(vault_version),
        check_ports=[vault_port],
        container_run_kwargs={
            "ports": {"8200/tcp": vault_port},
            "environment": {
                "VAULT_DEV_ROOT_TOKEN_ID": "testsecret",
                "VAULT_LOCAL_CONFIG": json.dumps(config),
            },
            "cap_add": "IPC_LOCK",
        },
        pull_before_start=True,
        skip_on_pull_failure=True,
        skip_if_docker_client_not_connectable=True,
    )
    with factory.started() as factory:
        attempts = 0
        while attempts < 3:
            attempts += 1
            time.sleep(1)
            ret = shell.run(
                VAULT_BINARY,
                "login",
                "token=testsecret",
                env={"VAULT_ADDR": "http://127.0.0.1:{}".format(vault_port)},
            )
            if ret.returncode == 0:
                break
            log.debug("Failed to authenticate against vault:\n%s", ret)
            time.sleep(4)
        else:
            pytest.fail("Failed to login to vault")

        ret = shell.run(
            VAULT_BINARY,
            "policy",
            "write",
            "testpolicy",
            "{}/vault.hcl".format(RUNTIME_VARS.FILES),
            env={"VAULT_ADDR": "http://127.0.0.1:{}".format(vault_port)},
        )
        if ret.returncode != 0:
            log.debug("Failed to assign policy to vault:\n%s", ret)
            pytest.fail("unable to assign policy to vault")
        yield vault_version


@pytest.fixture(scope="module")
def sys_mod(modules):
    return modules.sys


@pytest.fixture
def vault(loaders, modules, vault_container_version, shell, vault_port):
    try:
        yield modules.vault
    finally:
        # We're explicitly using the vault CLI and not the salt vault module
        secret_path = "secret/my"
        ret = shell.run(
            VAULT_BINARY,
            "kv",
            "list",
            "--format=json",
            secret_path,
            env={"VAULT_ADDR": "http://127.0.0.1:{}".format(vault_port)},
        )
        if ret.returncode == 0:
            for secret in ret.data:
                secret_path = "secret/my/{}".format(secret)
                ret = shell.run(
                    VAULT_BINARY,
                    "kv",
                    "delete",
                    secret_path,
                    env={"VAULT_ADDR": "http://127.0.0.1:{}".format(vault_port)},
                )
                ret = shell.run(
                    VAULT_BINARY,
                    "kv",
                    "metadata",
                    "delete",
                    secret_path,
                    env={"VAULT_ADDR": "http://127.0.0.1:{}".format(vault_port)},
                )


@pytest.mark.windows_whitelisted
def test_vault_read_secret_issue_61084(sys_mod):
    """
    Test issue 61084. `sys.argspec` should return valid data and not throw a
    TypeError due to pickling
    This should probably be a pre-commit check or something
    """
    result = sys_mod.argspec("vault.read_secret")
    assert isinstance(result, dict)
    assert isinstance(result.get("vault.read_secret"), dict)


@pytest.mark.windows_whitelisted
def test_vault_list_secrets_issue_61084(sys_mod):
    """
    Test issue 61084. `sys.argspec` should return valid data and not throw a
    TypeError due to pickling
    This should probably be a pre-commit check or something
    """
    result = sys_mod.argspec("vault.list_secrets")
    assert isinstance(result, dict)
    assert isinstance(result.get("vault.list_secrets"), dict)


def test_write_read_secret(vault, vault_container_version):
    ret = vault.write_secret(path="secret/my/secret", user="foo", password="bar")
    if vault_container_version == "0.9.6":
        assert ret is True
        ret = vault.read_secret(path="secret/my/secret")
        assert ret == {
            "password": "bar",
            "user": "foo",
        }
        ret = vault.read_secret(path="secret/my/secret", key="user")
        assert ret == "foo"
    else:
        # write_secret output:
        # {'created_time': '2020-01-12T23:09:34.571294241Z', 'destroyed': False,
        # 'version': 1, 'deletion_time': ''}
        assert ret
        expected_write = {"destroyed": False, "deletion_time": ""}
        for key in list(ret):
            if key not in expected_write:
                ret.pop(key)
        assert ret == expected_write

        ret = vault.read_secret("secret/my/secret", metadata=True)
        # read_secret output:
        # {'data': {'password': 'bar', 'user': 'foo'},
        # 'metadata': {'created_time': '2020-01-12T23:07:18.829326918Z', 'destroyed': False,
        # 'version': 1, 'deletion_time': ''}}
        assert ret
        assert "data" in ret
        expected_read = {"password": "bar", "user": "foo"}
        assert "metadata" in ret
        assert ret["data"] == expected_read

        ret = vault.read_secret("secret/my/secret")
        for key in list(ret):
            if key not in expected_read:
                ret.pop(key)
        assert ret == expected_read

        ret = vault.read_secret("secret/my/secret", key="user")
        assert ret == "foo"


def test_write_raw_read_secret(vault, vault_container_version):
    ret = vault.write_raw(
        "secret/my/secret2", raw={"user2": "foo2", "password2": "bar2"}
    )
    if vault_container_version == "0.9.6":
        assert ret is True
        ret = vault.read_secret("secret/my/secret2")
        assert ret == {
            "password2": "bar2",
            "user2": "foo2",
        }
    else:
        assert ret
        expected_write = {"destroyed": False, "deletion_time": ""}
        for key in list(ret):
            if key not in expected_write:
                ret.pop(key)
        assert ret == expected_write

        expected_read = {"password2": "bar2", "user2": "foo2"}
        ret = vault.read_secret("secret/my/secret2", metadata=True)
        assert ret
        assert "metadata" in ret
        assert "data" in ret
        assert ret["data"] == expected_read

        ret = vault.read_secret("secret/my/secret2")
        for key in list(ret):
            if key not in expected_read:
                ret.pop(key)
        assert ret == expected_read


@pytest.fixture
def existing_secret(vault, vault_container_version):
    ret = vault.write_secret("secret/my/secret", user="foo", password="bar")
    if vault_container_version == "0.9.6":
        assert ret is True
    else:
        expected_write = {"destroyed": False, "deletion_time": ""}
        for key in list(ret):
            if key not in expected_write:
                ret.pop(key)
        assert ret == expected_write


@pytest.mark.usefixtures("existing_secret")
def test_delete_secret(vault):
    ret = vault.delete_secret("secret/my/secret")
    assert ret is True


@pytest.mark.usefixtures("existing_secret")
def test_list_secrets(vault):
    ret = vault.list_secrets("secret/my/")
    assert ret
    assert "keys" in ret
    assert ret["keys"] == ["secret"]


@pytest.mark.usefixtures("existing_secret")
def test_destroy_secret_kv2(vault, vault_container_version):
    if vault_container_version == "0.9.6":
        pytest.skip("Test not applicable to vault=={}".format(vault_container_version))
    ret = vault.destroy_secret("secret/my/secret", "1")
    assert ret is True



import pytest

import salt.modules.win_servermanager as win_servermanager
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {win_servermanager: {"__grains__": {"osversion": "6.2"}}}


def test_install():
    """
    Test win_servermanager.install
    """
    mock_out = {
        "Success": True,
        "RestartNeeded": 1,
        "FeatureResult": [
            {
                "Id": 338,
                "Name": "XPS-Viewer",
                "DisplayName": "XPS Viewer",
                "Success": True,
                "RestartNeeded": False,
                "Message": "",
                "SkipReason": 0,
            }
        ],
        "ExitCode": 0,
    }
    expected = {
        "ExitCode": 0,
        "RestartNeeded": False,
        "Restarted": False,
        "Features": {
            "XPS-Viewer": {
                "DisplayName": "XPS Viewer",
                "Message": "",
                "RestartNeeded": False,
                "SkipReason": 0,
                "Success": True,
            }
        },
        "Success": True,
    }

    mock_reboot = MagicMock(return_value=True)
    with patch.object(
        win_servermanager, "_pshell_json", return_value=mock_out
    ), patch.dict(win_servermanager.__salt__, {"system.reboot": mock_reboot}):
        result = win_servermanager.install("XPS-Viewer")
        assert result == expected


def test_install_restart():
    """
    Test win_servermanager.install when restart=True
    """
    mock_out = {
        "Success": True,
        "RestartNeeded": 1,
        "FeatureResult": [
            {
                "Id": 338,
                "Name": "XPS-Viewer",
                "DisplayName": "XPS Viewer",
                "Success": True,
                "RestartNeeded": True,
                "Message": "",
                "SkipReason": 0,
            }
        ],
        "ExitCode": 0,
    }
    expected = {
        "ExitCode": 0,
        "RestartNeeded": True,
        "Restarted": True,
        "Features": {
            "XPS-Viewer": {
                "DisplayName": "XPS Viewer",
                "Message": "",
                "RestartNeeded": True,
                "SkipReason": 0,
                "Success": True,
            }
        },
        "Success": True,
    }

    mock_reboot = MagicMock(return_value=True)
    with patch.object(
        win_servermanager, "_pshell_json", return_value=mock_out
    ), patch.dict(win_servermanager.__salt__, {"system.reboot": mock_reboot}):
        result = win_servermanager.install("XPS-Viewer", restart=True)
        mock_reboot.assert_called_once()
        assert result == expected



import pytest

from tests.support.case import ModuleCase
from tests.support.mixins import SaltReturnAssertsMixin


@pytest.mark.windows_whitelisted
class PublishModuleTest(ModuleCase, SaltReturnAssertsMixin):
    """
    Validate the publish module
    """

    @pytest.mark.slow_test
    def test_publish(self):
        """
        publish.publish
        """
        ret = self.run_function(
            "publish.publish", ["minion", "test.ping"], f_timeout=50
        )
        self.assertEqual(ret, {"minion": True})

        ret = self.run_function(
            "publish.publish",
            ["minion", "test.kwarg"],
            f_arg="cheese=spam",
            f_timeout=50,
        )
        ret = ret["minion"]

        check_true = (
            "cheese",
            "__pub_arg",
            "__pub_fun",
            "__pub_id",
            "__pub_jid",
            "__pub_ret",
            "__pub_tgt",
            "__pub_tgt_type",
        )
        for name in check_true:
            if name not in ret:
                print(name)
            self.assertTrue(name in ret)

        self.assertEqual(ret["cheese"], "spam")
        self.assertEqual(ret["__pub_arg"], [{"cheese": "spam"}])
        self.assertEqual(ret["__pub_id"], "minion")
        self.assertEqual(ret["__pub_fun"], "test.kwarg")

    @pytest.mark.slow_test
    def test_publish_yaml_args(self):
        """
        test publish.publish yaml args formatting
        """
        ret = self.run_function(
            "publish.publish", ["minion", "test.ping"], f_timeout=50
        )
        self.assertEqual(ret, {"minion": True})

        test_args_list = ["saltines, si", "crackers, nein", "cheese, indeed"]
        test_args = '["{args[0]}", "{args[1]}", "{args[2]}"]'.format(
            args=test_args_list
        )
        ret = self.run_function(
            "publish.publish", ["minion", "test.arg", test_args], f_timeout=50
        )
        ret = ret["minion"]

        check_true = (
            "__pub_arg",
            "__pub_fun",
            "__pub_id",
            "__pub_jid",
            "__pub_ret",
            "__pub_tgt",
            "__pub_tgt_type",
        )
        for name in check_true:
            if name not in ret["kwargs"]:
                print(name)
            self.assertTrue(name in ret["kwargs"])

        self.assertEqual(ret["args"], test_args_list)
        self.assertEqual(ret["kwargs"]["__pub_id"], "minion")
        self.assertEqual(ret["kwargs"]["__pub_fun"], "test.arg")

    @pytest.mark.slow_test
    def test_full_data(self):
        """
        publish.full_data
        """
        ret = self.run_function(
            "publish.full_data", ["minion", "test.fib", 20], f_timeout=50
        )
        self.assertTrue(ret)
        self.assertEqual(ret["minion"]["ret"][0], 6765)

    @pytest.mark.slow_test
    def test_kwarg(self):
        """
        Verify that the pub data is making it to the minion functions
        """
        ret = self.run_function(
            "publish.full_data",
            ["minion", "test.kwarg"],
            f_arg="cheese=spam",
            f_timeout=50,
        )
        ret = ret["minion"]["ret"]

        check_true = (
            "cheese",
            "__pub_arg",
            "__pub_fun",
            "__pub_id",
            "__pub_jid",
            "__pub_ret",
            "__pub_tgt",
            "__pub_tgt_type",
        )
        for name in check_true:
            if name not in ret:
                print(name)
            self.assertTrue(name in ret)

        self.assertEqual(ret["cheese"], "spam")
        self.assertEqual(ret["__pub_arg"], [{"cheese": "spam"}])
        self.assertEqual(ret["__pub_id"], "minion")
        self.assertEqual(ret["__pub_fun"], "test.kwarg")

        ret = self.run_function(
            "publish.full_data", ["minion", "test.kwarg"], cheese="spam", f_timeout=50
        )
        self.assertIn("The following keyword arguments are not valid", ret)

    @pytest.mark.slow_test
    def test_reject_minion(self):
        """
        Test bad authentication
        """
        ret = self.run_function(
            "publish.publish", ["minion", "cmd.run", ["echo foo"]], f_timeout=50
        )
        self.assertEqual(ret, {})



import logging
import os
import threading
import time

import pytest

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.windows_whitelisted,
]

log = logging.getLogger(__name__)


def minion_func(salt_minion, event_listener, salt_master, timeout):
    start = time.time()
    with salt_minion.started(start_timeout=timeout * 2, max_start_attempts=1):
        new_start = time.time()
        while time.time() < new_start + (timeout * 2):
            if event_listener.get_events(
                [(salt_master.id, f"salt/job/*/ret/{salt_minion.id}")],
                after_time=start,
            ):
                break
            time.sleep(5)


@pytest.fixture(scope="module")
def timeout():
    return int(os.environ.get("SALT_CI_REAUTH_MASTER_WAIT", 150))


def test_reauth(salt_cli, salt_minion, salt_master, timeout, event_listener):
    # Make sure they can communicate
    assert salt_cli.run("test.ping", minion_tgt=salt_minion.id).data is True
    # Stop the master and minion
    salt_master.terminate()
    salt_minion.terminate()
    log.debug(
        "Master and minion stopped for reauth test, waiting for %s seconds", timeout
    )
    log.debug("Restarting the reauth minion")

    # We need to have the minion attempting to start in a different process
    # when we try to start the master
    minion_proc = threading.Thread(
        target=minion_func, args=(salt_minion, event_listener, salt_master, timeout)
    )
    minion_proc.start()
    time.sleep(timeout)
    log.debug("Restarting the reauth master")
    start = time.time()
    salt_master.start()
    event_listener.wait_for_events(
        [(salt_master.id, f"salt/minion/{salt_minion.id}/start")],
        after_time=start,
        timeout=timeout * 2,
    )
    assert salt_cli.run("test.ping", minion_tgt=salt_minion.id).data is True
    minion_proc.join()



"""
Integration tests for modules/useradd.py and modules/win_useradd.py
"""
import pytest
from saltfactories.utils import random_string

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
    pytest.mark.slow_test,
]


@pytest.fixture(scope="function")
def setup_teardown_vars(salt_call_cli):
    user_name = random_string("RS-", lowercase=False)
    group_name = random_string("RS-", lowercase=False)
    try:
        yield user_name, group_name
    finally:
        salt_call_cli.run("user.delete", user_name, True, True)
        salt_call_cli.run("group.delete", group_name)
        salt_call_cli.run("lgpo.set", "computer_policy={'Minimum Password Length': 0}")


@pytest.mark.skip_on_windows(reason="Windows does not do user checks")
@pytest.mark.destructive_test
@pytest.mark.skip_if_not_root
def test_groups_includes_primary(setup_teardown_vars, grains, salt_call_cli):
    # Let's create a user, which usually creates the group matching the
    # name
    uname = random_string("RS-", lowercase=False)
    ret = salt_call_cli.run("user.add", uname)
    if ret.data is False:
        # Skip because creating is not what we're testing here
        salt_call_cli.run("user.delete", [uname, True, True])
        pytest.skip("Failed to create user")

    try:
        uinfo = salt_call_cli.run("user.info", uname)
        if grains["os_family"] in ("Suse",):
            assert "users" in uinfo.data["groups"]
        else:
            assert uname in uinfo.data["groups"]

        # This uid is available, store it
        uid = uinfo.data["uid"]

        salt_call_cli.run("user.delete", uname, True, True)

        # Now, a weird group id
        gname = random_string("RS-", lowercase=False)
        ret = salt_call_cli.run("group.add", gname)
        if ret.data is False:
            salt_call_cli.run("group.delete", gname, True, True)
            pytest.skip("Failed to create group")

        ginfo = salt_call_cli.run("group.info", gname)
        ginfo = ginfo.data

        # And create the user with that gid
        ret = salt_call_cli.run("user.add", uname, uid, ginfo["gid"])
        if ret.data is False:
            # Skip because creating is not what we're testing here
            salt_call_cli.run("user.delete", [uname, True, True])
            pytest.skip("Failed to create user")

        uinfo = salt_call_cli.run("user.info", uname)
        assert gname in uinfo.data["groups"]

    except AssertionError:
        pytest.raises(salt_call_cli.run("user.delete", [uname, True, True]))


@pytest.mark.skip_on_windows(reason="Windows does not do user checks")
@pytest.mark.destructive_test
@pytest.mark.skip_if_not_root
def test_user_primary_group(setup_teardown_vars, salt_call_cli):
    """
    Tests the primary_group function
    """
    name = "saltyuser"

    # Create a user to test primary group function
    ret = salt_call_cli.run("user.add", name)
    if ret.data is False:
        salt_call_cli.run("user.delete", name)
        pytest.skip("Failed to create a user")

    # Test useradd.primary_group
    primary_group = salt_call_cli.run("user.primary_group", name)
    uid_info = salt_call_cli.run("user.info", name)
    assert primary_group.data in uid_info.data["groups"]


@pytest.mark.skip_unless_on_windows(reason="Test is only applicable to Windows.")
@pytest.mark.destructive_test
def test_add_user(setup_teardown_vars, salt_call_cli):
    """
    Test adding a user.
    """
    user_name = setup_teardown_vars[0]
    salt_call_cli.run("user.add", user_name)
    user_add = salt_call_cli.run("user.list_users")
    assert user_name in user_add.data


@pytest.mark.skip_unless_on_windows(reason="Test is only applicable to Windows.")
@pytest.mark.destructive_test
def test_add_group(setup_teardown_vars, salt_call_cli):
    """
    Test adding a user and check its apart of a group.
    """
    group_name = setup_teardown_vars[1]
    salt_call_cli.run("group.add", group_name)
    group_list = salt_call_cli.run("group.list_groups")
    assert group_name in group_list.data


@pytest.mark.skip_unless_on_windows(reason="Test is only applicable to Windows.")
@pytest.mark.destructive_test
def test_add_user_to_group(setup_teardown_vars, salt_call_cli):
    """
    Test adding a user to a group.
    """
    user_name = setup_teardown_vars[0]
    group_name = setup_teardown_vars[1]

    salt_call_cli.run("group.add", group_name)
    # And create the user as a member of that group
    salt_call_cli.run("user.add", user_name, groups=group_name)

    user_info = salt_call_cli.run("user.info", user_name)
    assert group_name in user_info.data["groups"]


@pytest.mark.skip_unless_on_windows(reason="Test is only applicable to Windows.")
@pytest.mark.destructive_test
def test_add_user_addgroup(setup_teardown_vars, salt_call_cli):
    """
    Test adding a user to a group with groupadd.
    """
    user_name = setup_teardown_vars[0]
    group_name = setup_teardown_vars[1]

    salt_call_cli.run("group.add", group_name)
    salt_call_cli.run("user.add", user_name)

    salt_call_cli.run("user.addgroup", user_name, group_name)
    info = salt_call_cli.run("user.info", user_name)
    assert [group_name] == info.data["groups"]


@pytest.mark.skip_unless_on_windows(reason="Test is only applicable to Windows.")
@pytest.mark.destructive_test
def test_user_chhome(setup_teardown_vars, salt_call_cli):
    """
    Test changing a users home dir.
    """
    user_dir = r"c:\salt"
    user_name = setup_teardown_vars[0]
    salt_call_cli.run("user.add", user_name)
    salt_call_cli.run("user.chhome", user_name, user_dir)

    info = salt_call_cli.run("user.info", user_name)
    assert user_dir == info.data["home"]


@pytest.mark.skip_unless_on_windows(reason="Test is only applicable to Windows.")
@pytest.mark.destructive_test
def test_user_chprofile(setup_teardown_vars, salt_call_cli):
    """
    Test changing a users profile.
    """
    config = r"c:\salt\config"
    user_name = setup_teardown_vars[0]
    salt_call_cli.run("user.add", user_name)

    salt_call_cli.run("user.chprofile", user_name, config)
    info = salt_call_cli.run("user.info", user_name)
    assert config == info.data["profile"]


@pytest.mark.skip_unless_on_windows(reason="Test is only applicable to Windows.")
@pytest.mark.destructive_test
def test_user_chfullname(setup_teardown_vars, salt_call_cli):
    """
    Test changing a users fullname.
    """
    name = "Salt Test"
    user_name = setup_teardown_vars[0]
    salt_call_cli.run("user.add", user_name)

    salt_call_cli.run("user.chfullname", user_name, name)
    info = salt_call_cli.run("user.info", user_name)
    assert name == info.data["fullname"]


@pytest.mark.skip_unless_on_windows(reason="Test is only applicable to Windows.")
@pytest.mark.destructive_test
def test_user_delete(setup_teardown_vars, salt_call_cli):
    """
    Test deleting a user.
    """
    user_name = setup_teardown_vars[0]
    salt_call_cli.run("user.add", user_name)
    salt_call_cli.run("user.delete", user_name)
    ret = salt_call_cli.run("user.info", user_name)
    assert {} == ret.data


@pytest.mark.skip_unless_on_windows(reason="Test is only applicable to Windows.")
@pytest.mark.destructive_test
def test_user_removegroup(setup_teardown_vars, salt_call_cli):
    """
    Test removing a group.
    """
    user_name = setup_teardown_vars[0]
    group_name = setup_teardown_vars[1]

    salt_call_cli.run("user.add", user_name)
    salt_call_cli.run("group.add", group_name)

    salt_call_cli.run("user.addgroup", user_name, group_name)
    ret = salt_call_cli.run("user.list_groups", user_name)
    assert [group_name] == ret.data

    salt_call_cli.run("user.removegroup", user_name, group_name)
    ret = salt_call_cli.run("user.list_groups", user_name)
    assert [group_name] not in ret.data


@pytest.mark.skip_unless_on_windows(reason="Test is only applicable to Windows.")
@pytest.mark.destructive_test
def test_user_rename(setup_teardown_vars, salt_call_cli):
    """
    Test changing a users name.
    """
    name = "newuser"
    user_name = setup_teardown_vars[0]
    salt_call_cli.run("user.add", user_name)

    salt_call_cli.run("user.rename", user_name, name)
    info = salt_call_cli.run("user.info", name)

    assert info.data["active"] is True


@pytest.mark.skip_unless_on_windows(reason="Test is only applicable to Windows.")
@pytest.mark.destructive_test
def test_user_setpassword(setup_teardown_vars, salt_call_cli):
    """
    Test setting a password.
    """
    passwd = "sup3rs3cr3T!"
    user_name = setup_teardown_vars[0]

    salt_call_cli.run("user.add", user_name)
    ret = salt_call_cli.run("user.setpassword", user_name, passwd)
    assert ret.data is True


@pytest.mark.skip_unless_on_windows(reason="Test is only applicable to Windows.")
@pytest.mark.destructive_test
def test_user_setpassword_policy(setup_teardown_vars, salt_call_cli):
    """
    Test setting a password with a password policy.
    """
    passwd = "test"
    user_name = setup_teardown_vars[0]

    # attempt to set a password policy that will cause a failure when creating a user
    salt_call_cli.run("lgpo.set", "computer_policy={'Minimum Password Length': 8}")
    ret = salt_call_cli.run("user.add", user_name, password=passwd)

    # fix the policy and store the previous strerror in ret to cleanup
    salt_call_cli.run("lgpo.set", "computer_policy={'Minimum Password Length': 0}")
    assert (
        ret.data == "The password does not meet the password policy requirements."
        " Check the minimum password length, password complexity and"
        " password history requirements."
    )



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.pagerduty as pagerduty
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {pagerduty: {}}


def test_create_event():
    """
    Test to create an event on the PagerDuty service.
    """
    name = "This is a server warning message"
    details = "This is a much more detailed message"
    service_key = "9abcd123456789efabcde362783cdbaf"
    profile = "my-pagerduty-account"

    ret = {"name": name, "result": None, "comment": "", "changes": {}}

    with patch.dict(pagerduty.__opts__, {"test": True}):
        comt = "Need to create event: {}".format(name)
        ret.update({"comment": comt})
        assert pagerduty.create_event(name, details, service_key, profile) == ret

    with patch.dict(pagerduty.__opts__, {"test": False}):
        mock_t = MagicMock(return_value=True)
        with patch.dict(pagerduty.__salt__, {"pagerduty.create_event": mock_t}):
            comt = "Created event: {}".format(name)
            ret.update({"comment": comt, "result": True})
            assert pagerduty.create_event(name, details, service_key, profile) == ret



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.lvs_service as lvs_service
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {lvs_service: {}}


def test_present():
    """
    Test to ensure that the named service is present.
    """
    name = "lvsrs"

    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    mock_check = MagicMock(
        side_effect=[
            True,
            True,
            True,
            False,
            True,
            False,
            True,
            False,
            False,
            False,
            False,
        ]
    )
    mock_edit = MagicMock(side_effect=[True, False])
    mock_add = MagicMock(side_effect=[True, False])
    with patch.dict(
        lvs_service.__salt__,
        {
            "lvs.check_service": mock_check,
            "lvs.edit_service": mock_edit,
            "lvs.add_service": mock_add,
        },
    ):
        with patch.dict(lvs_service.__opts__, {"test": True}):
            comt = "LVS Service lvsrs is present"
            ret.update({"comment": comt})
            assert lvs_service.present(name) == ret

            comt = "LVS Service lvsrs is present but some options should update"
            ret.update({"comment": comt, "result": None})
            assert lvs_service.present(name) == ret

        with patch.dict(lvs_service.__opts__, {"test": False}):
            comt = "LVS Service lvsrs has been updated"
            ret.update(
                {"comment": comt, "result": True, "changes": {"lvsrs": "Update"}}
            )
            assert lvs_service.present(name) == ret

            comt = "LVS Service lvsrs update failed"
            ret.update({"comment": comt, "result": False, "changes": {}})
            assert lvs_service.present(name) == ret

        with patch.dict(lvs_service.__opts__, {"test": True}):
            comt = "LVS Service lvsrs is not present and needs to be created"
            ret.update({"comment": comt, "result": None})
            assert lvs_service.present(name) == ret

        with patch.dict(lvs_service.__opts__, {"test": False}):
            comt = "LVS Service lvsrs has been created"
            ret.update(
                {"comment": comt, "result": True, "changes": {"lvsrs": "Present"}}
            )
            assert lvs_service.present(name) == ret

            comt = "LVS Service lvsrs create failed(False)"
            ret.update({"comment": comt, "result": False, "changes": {}})
            assert lvs_service.present(name) == ret


def test_absent():
    """
    Test to ensure the LVS Real Server in specified service is absent.
    """
    name = "lvsrs"

    ret = {"name": name, "result": None, "comment": "", "changes": {}}

    mock_check = MagicMock(side_effect=[True, True, True, False])
    mock_delete = MagicMock(side_effect=[True, False])
    with patch.dict(
        lvs_service.__salt__,
        {"lvs.check_service": mock_check, "lvs.delete_service": mock_delete},
    ):
        with patch.dict(lvs_service.__opts__, {"test": True}):
            comt = "LVS Service lvsrs is present and needs to be removed"
            ret.update({"comment": comt})
            assert lvs_service.absent(name) == ret

        with patch.dict(lvs_service.__opts__, {"test": False}):
            comt = "LVS Service lvsrs has been removed"
            ret.update(
                {"comment": comt, "result": True, "changes": {"lvsrs": "Absent"}}
            )
            assert lvs_service.absent(name) == ret

            comt = "LVS Service lvsrs removed failed(False)"
            ret.update({"comment": comt, "result": False, "changes": {}})
            assert lvs_service.absent(name) == ret

        comt = "LVS Service lvsrs is not present, so it cannot be removed"
        ret.update({"comment": comt, "result": True})
        assert lvs_service.absent(name) == ret



"""
Grains from cloud metadata servers at 169.254.169.254 in
google compute engine

.. versionadded:: 3005

:depends: requests

To enable these grains that pull from the http://169.254.169.254/computeMetadata/v1/
metadata server set `metadata_server_grains: True` in the minion config.

.. code-block:: yaml

    metadata_server_grains: True

"""

import logging

import salt.utils.http as http
import salt.utils.json

HOST = "http://169.254.169.254"
URL = f"{HOST}/computeMetadata/v1/?alt=json&recursive=true"
log = logging.getLogger(__name__)


def __virtual__():
    # Check if metadata_server_grains minion option is enabled
    if __opts__.get("metadata_server_grains", False) is False:
        return False
    googletest = http.query(HOST, status=True, headers=True)
    if (
        googletest.get("status", 404) != 200
        or googletest.get("headers", {}).get("Metadata-Flavor", False) != "Google"
    ):
        return False
    return True


def metadata():
    """Takes no arguments, returns a dictionary of metadata values from Google."""
    log.debug("All checks true - loading gce metadata")
    result = http.query(URL, headers=True, header_list=["Metadata-Flavor: Google"])
    metadata = salt.utils.json.loads(result.get("body", {}))

    return metadata



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.keyboard as keyboard
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {keyboard: {}}


# 'system' function tests: 1


def test_system():
    """
    Test to set the keyboard layout for the system.
    """
    name = "salt"

    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    mock = MagicMock(side_effect=[name, "", "", ""])
    mock_t = MagicMock(side_effect=[True, False])
    with patch.dict(
        keyboard.__salt__, {"keyboard.get_sys": mock, "keyboard.set_sys": mock_t}
    ):
        comt = "System layout {} already set".format(name)
        ret.update({"comment": comt})
        assert keyboard.system(name) == ret

        with patch.dict(keyboard.__opts__, {"test": True}):
            comt = "System layout {} needs to be set".format(name)
            ret.update({"comment": comt, "result": None})
            assert keyboard.system(name) == ret

        with patch.dict(keyboard.__opts__, {"test": False}):
            comt = "Set system keyboard layout {}".format(name)
            ret.update({"comment": comt, "result": True, "changes": {"layout": name}})
            assert keyboard.system(name) == ret

            comt = "Failed to set system keyboard layout"
            ret.update({"comment": comt, "result": False, "changes": {}})
            assert keyboard.system(name) == ret


# 'xorg' function tests: 1


def test_xorg():
    """
    Test to set the keyboard layout for XOrg.
    """
    name = "salt"

    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    mock = MagicMock(side_effect=[name, "", "", ""])
    mock_t = MagicMock(side_effect=[True, False])
    with patch.dict(
        keyboard.__salt__, {"keyboard.get_x": mock, "keyboard.set_x": mock_t}
    ):
        comt = "XOrg layout {} already set".format(name)
        ret.update({"comment": comt})
        assert keyboard.xorg(name) == ret

        with patch.dict(keyboard.__opts__, {"test": True}):
            comt = "XOrg layout {} needs to be set".format(name)
            ret.update({"comment": comt, "result": None})
            assert keyboard.xorg(name) == ret

        with patch.dict(keyboard.__opts__, {"test": False}):
            comt = "Set XOrg keyboard layout {}".format(name)
            ret.update({"comment": comt, "result": True, "changes": {"layout": name}})
            assert keyboard.xorg(name) == ret

            comt = "Failed to set XOrg keyboard layout"
            ret.update({"comment": comt, "result": False, "changes": {}})
            assert keyboard.xorg(name) == ret



import logging

import pytest

from tests.support.mock import MagicMock, patch
from tests.support.pytest.transport import PubServerChannelProcess

log = logging.getLogger(__name__)


pytestmark = [
    pytest.mark.skip_on_freebsd(reason="Temporarily skipped on FreeBSD."),
    pytest.mark.skip_on_spawning_platform(
        reason="These tests are currently broken on spawning platforms. Need to be rewritten.",
    ),
]


@pytest.mark.skip_on_windows
@pytest.mark.slow_test
def test_zeromq_filtering(salt_master, salt_minion):
    """
    Test sending messages to publisher using UDP with zeromq_filtering enabled
    """
    opts = dict(
        salt_master.config.copy(),
        ipc_mode="ipc",
        pub_hwm=0,
        zmq_filtering=True,
        acceptance_wait_time=5,
    )
    send_num = 1
    expect = []
    with patch(
        "salt.utils.minions.CkMinions.check_minions",
        MagicMock(
            return_value={
                "minions": [salt_minion.id],
                "missing": [],
                "ssh_minions": False,
            }
        ),
    ):
        with PubServerChannelProcess(
            opts, salt_minion.config.copy(), zmq_filtering=True
        ) as server_channel:
            expect.append(send_num)
            load = {"tgt_type": "glob", "tgt": "*", "jid": send_num}
            server_channel.publish(load)
        results = server_channel.collector.results
        assert len(results) == send_num, "{} != {}, difference: {}".format(
            len(results), send_num, set(expect).difference(results)
        )
